"""
:math:`LLAB(l:c)` Colour Appearance Model
=========================================

Defines the *:math:`LLAB(l:c)`* colour appearance model objects:

-   :class:`colour.appearance.InductionFactors_LLAB`
-   :attr:`colour.VIEWING_CONDITIONS_LLAB`
-   :class:`colour.CAM_Specification_LLAB`
-   :func:`colour.XYZ_to_LLAB`

References
----------
-   :cite:`Fairchild2013x` : Fairchild, M. D. (2013). LLAB Model. In Color
    Appearance Models (3rd ed., pp. 6025-6178). Wiley. ISBN:B00DAYO8E2
-   :cite:`Luo1996b` : Luo, Ming Ronnier, Lo, M.-C., & Kuo, W.-G. (1996). The
    LLAB (l:c) colour model. Color Research & Application, 21(6), 412-429.
    doi:10.1002/(SICI)1520-6378(199612)21:6<412::AID-COL4>3.0.CO;2-Z
-   :cite:`Luo1996c` : Luo, Ming Ronnier, & Morovic, J. (1996). Two Unsolved
    Issues in Colour Management - Colour Appearance and Gamut Mapping.
    Conference: 5th International Conference on High Technology: Imaging
    Science and Technology - Evolution & Promise, 136-147.
    http://www.researchgate.net/publication/\
236348295_Two_Unsolved_Issues_in_Colour_Management__\
Colour_Appearance_and_Gamut_Mapping
"""

import numpy as np
from collections import namedtuple
from dataclasses import dataclass, field

from colour.algebra import polar_to_cartesian, spow, vector_dot
from colour.hints import (
    ArrayLike,
    FloatingOrArrayLike,
    FloatingOrNDArray,
    NDArray,
    Optional,
)
from colour.utilities import (
    CaseInsensitiveMapping,
    MixinDataclassArithmetic,
    as_float,
    as_float_array,
    from_range_degrees,
    to_domain_100,
    tsplit,
    tstack,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "InductionFactors_LLAB",
    "VIEWING_CONDITIONS_LLAB",
    "MATRIX_XYZ_TO_RGB_LLAB",
    "MATRIX_RGB_TO_XYZ_LLAB",
    "CAM_ReferenceSpecification_LLAB",
    "CAM_Specification_LLAB",
    "XYZ_to_LLAB",
    "XYZ_to_RGB_LLAB",
    "chromatic_adaptation",
    "f",
    "opponent_colour_dimensions",
    "hue_angle",
    "chroma_correlate",
    "colourfulness_correlate",
    "saturation_correlate",
    "final_opponent_signals",
]


class InductionFactors_LLAB(
    namedtuple("InductionFactors_LLAB", ("D", "F_S", "F_L", "F_C"))
):
    """
    *:math:`LLAB(l:c)`* colour appearance model induction factors.

    Parameters
    ----------
    D
         *Discounting-the-Illuminant* factor :math:`D`.
    F_S
        Surround induction factor :math:`F_S`.
    F_L
        *Lightness* induction factor :math:`F_L`.
    F_C
        *Chroma* induction factor :math:`F_C`.

    References
    ----------
    :cite:`Fairchild2013x`, :cite:`Luo1996b`, :cite:`Luo1996c`
    """


VIEWING_CONDITIONS_LLAB: CaseInsensitiveMapping = CaseInsensitiveMapping(
    {
        "Reference Samples & Images, Average Surround, Subtending > 4": (
            InductionFactors_LLAB(1, 3, 0, 1)
        ),
        "Reference Samples & Images, Average Surround, Subtending < 4": (
            InductionFactors_LLAB(1, 3, 1, 1)
        ),
        "Television & VDU Displays, Dim Surround": (
            InductionFactors_LLAB(0.7, 3.5, 1, 1)
        ),
        "Cut Sheet Transparency, Dim Surround": (
            InductionFactors_LLAB(1, 5, 1, 1.1)
        ),
        "35mm Projection Transparency, Dark Surround": (
            InductionFactors_LLAB(0.7, 4, 1, 1)
        ),
    }
)
VIEWING_CONDITIONS_LLAB.__doc__ = """
Reference :math:`LLAB(l:c)` colour appearance model viewing conditions.

References
----------
:cite:`Fairchild2013x`, :cite:`Luo1996b`, :cite:`Luo1996c`

Aliases:

-   'ref_average_4_plus':
    'Reference Samples & Images, Average Surround, Subtending > 4'
-   'ref_average_4_minus':
    'Reference Samples & Images, Average Surround, Subtending < 4'
-   'tv_dim': 'Television & VDU Displays, Dim Surround'
-   'sheet_dim': 'Cut Sheet Transparency, Dim Surround'
-   'projected_dark': '35mm Projection Transparency, Dark Surround'
"""
VIEWING_CONDITIONS_LLAB["ref_average_4_plus"] = VIEWING_CONDITIONS_LLAB[
    "Reference Samples & Images, " "Average Surround, Subtending > 4"
]
VIEWING_CONDITIONS_LLAB["ref_average_4_minus"] = VIEWING_CONDITIONS_LLAB[
    "Reference Samples & Images, " "Average Surround, Subtending < 4"
]
VIEWING_CONDITIONS_LLAB["tv_dim"] = VIEWING_CONDITIONS_LLAB[
    "Television & VDU Displays, Dim Surround"
]
VIEWING_CONDITIONS_LLAB["sheet_dim"] = VIEWING_CONDITIONS_LLAB[
    "Cut Sheet Transparency, Dim Surround"
]
VIEWING_CONDITIONS_LLAB["projected_dark"] = VIEWING_CONDITIONS_LLAB[
    "35mm Projection Transparency, Dark Surround"
]

MATRIX_XYZ_TO_RGB_LLAB: NDArray = np.array(
    [
        [0.8951, 0.2664, -0.1614],
        [-0.7502, 1.7135, 0.0367],
        [0.0389, -0.0685, 1.0296],
    ]
)
"""
LLAB(l:c) colour appearance model *CIE XYZ* tristimulus values to normalised
cone responses matrix.
"""

MATRIX_RGB_TO_XYZ_LLAB: NDArray = np.linalg.inv(MATRIX_XYZ_TO_RGB_LLAB)
"""
LLAB(l:c) colour appearance model normalised cone responses to *CIE XYZ*
tristimulus values matrix.
"""


@dataclass
class CAM_ReferenceSpecification_LLAB(MixinDataclassArithmetic):
    """
    Define the *:math:`LLAB(l:c)`* colour appearance model reference
    specification.

    This specification has field names consistent with *Fairchild (2013)*
    reference.

    Parameters
    ----------
    L_L
        Correlate of *Lightness* :math:`L_L`.
    Ch_L
        Correlate of *chroma* :math:`Ch_L`.
    h_L
        *Hue* angle :math:`h_L` in degrees.
    s_L
        Correlate of *saturation* :math:`s_L`.
    C_L
        Correlate of *colourfulness* :math:`C_L`.
    HC
        *Hue* :math:`h` composition :math:`H^C`.
    A_L
        Opponent signal :math:`A_L`.
    B_L
        Opponent signal :math:`B_L`.

    References
    ----------
    :cite:`Fairchild2013x`, :cite:`Luo1996b`, :cite:`Luo1996c`
    """

    L_L: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    Ch_L: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    h_L: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    s_L: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    C_L: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    HC: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    A_L: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    B_L: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)


@dataclass
class CAM_Specification_LLAB(MixinDataclassArithmetic):
    """
    Define the *:math:`LLAB(l:c)`* colour appearance model specification.

    This specification has field names consistent with the remaining colour
    appearance models in :mod:`colour.appearance` but diverge from
    *Fairchild (2013)* reference.

    Parameters
    ----------
    J
        Correlate of *Lightness* :math:`L_L`.
    C
        Correlate of *chroma* :math:`Ch_L`.
    h
        *Hue* angle :math:`h_L` in degrees.
    s
        Correlate of *saturation* :math:`s_L`.
    M
        Correlate of *colourfulness* :math:`C_L`.
    HC
        *Hue* :math:`h` composition :math:`H^C`.
    a
        Opponent signal :math:`A_L`.
    b
        Opponent signal :math:`B_L`.

    Notes
    -----
    -   This specification is the one used in the current model implementation.

    References
    ----------
    :cite:`Fairchild2013x`, :cite:`Luo1996b`, :cite:`Luo1996c`
    """

    J: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    C: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    h: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    s: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    M: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    HC: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    a: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    b: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)


def XYZ_to_LLAB(
    XYZ: ArrayLike,
    XYZ_0: ArrayLike,
    Y_b: FloatingOrArrayLike,
    L: FloatingOrArrayLike,
    surround: InductionFactors_LLAB = VIEWING_CONDITIONS_LLAB[
        "Reference Samples & Images, Average Surround, Subtending < 4"
    ],
) -> CAM_Specification_LLAB:
    """
    Compute the *:math:`LLAB(l:c)`* colour appearance model correlates.

    Parameters
    ----------
    XYZ
        *CIE XYZ* tristimulus values of test sample / stimulus.
    XYZ_0
        *CIE XYZ* tristimulus values of reference white.
    Y_b
        Luminance factor of the background in :math:`cd/m^2`.
    L
        Absolute luminance :math:`L` of reference white in :math:`cd/m^2`.
    surround
         Surround viewing conditions induction factors.

    Returns
    -------
    :class:`colour.CAM_Specification_LLAB`
        *:math:`LLAB(l:c)`* colour appearance model specification.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``XYZ``    | [0, 100]              | [0, 1]        |
    +------------+-----------------------+---------------+
    | ``XYZ_0``  | [0, 100]              | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------------------------+-----------------------+---------------+
    | **Range**                    | **Scale - Reference** | **Scale - 1** |
    +==============================+=======================+===============+
    | ``CAM_Specification_LLAB.h`` | [0, 360]              | [0, 1]        |
    +------------------------------+-----------------------+---------------+

    References
    ----------
    :cite:`Fairchild2013x`, :cite:`Luo1996b`, :cite:`Luo1996c`

    Examples
    --------
    >>> XYZ = np.array([19.01, 20.00, 21.78])
    >>> XYZ_0 = np.array([95.05, 100.00, 108.88])
    >>> Y_b = 20.0
    >>> L = 318.31
    >>> surround = VIEWING_CONDITIONS_LLAB['ref_average_4_minus']
    >>> XYZ_to_LLAB(XYZ, XYZ_0, Y_b, L, surround)  # doctest: +ELLIPSIS
    CAM_Specification_LLAB(J=37.3668650..., C=0.0089496..., h=270..., \
s=0.0002395..., M=0.0190185..., HC=None, a=..., b=-0.0190185...)
    """

    _X, Y, _Z = tsplit(to_domain_100(XYZ))
    RGB = XYZ_to_RGB_LLAB(to_domain_100(XYZ))
    RGB_0 = XYZ_to_RGB_LLAB(to_domain_100(XYZ_0))

    # Reference illuminant *CIE Standard Illuminant D Series* *D65*.
    XYZ_0r = np.array([95.05, 100.00, 108.88])
    RGB_0r = XYZ_to_RGB_LLAB(XYZ_0r)

    # Computing chromatic adaptation.
    XYZ_r = chromatic_adaptation(RGB, RGB_0, RGB_0r, Y, surround.D)

    # -------------------------------------------------------------------------
    # Computing the correlate of *Lightness* :math:`L_L`.
    # -------------------------------------------------------------------------
    # Computing opponent colour dimensions.
    L_L, a, b = tsplit(
        opponent_colour_dimensions(XYZ_r, Y_b, surround.F_S, surround.F_L)
    )

    # Computing perceptual correlates.
    # -------------------------------------------------------------------------
    # Computing the correlate of *chroma* :math:`Ch_L`.
    # -------------------------------------------------------------------------
    Ch_L = chroma_correlate(a, b)

    # -------------------------------------------------------------------------
    # Computing the correlate of *colourfulness* :math:`C_L`.
    # -------------------------------------------------------------------------
    C_L = colourfulness_correlate(L, L_L, Ch_L, surround.F_C)

    # -------------------------------------------------------------------------
    # Computing the correlate of *saturation* :math:`s_L`.
    # -------------------------------------------------------------------------
    s_L = saturation_correlate(Ch_L, L_L)

    # -------------------------------------------------------------------------
    # Computing the *hue* angle :math:`h_L`.
    # -------------------------------------------------------------------------
    h_L = hue_angle(a, b)
    # TODO: Implement hue composition computation.

    # -------------------------------------------------------------------------
    # Computing final opponent signals.
    # -------------------------------------------------------------------------
    A_L, B_L = tsplit(final_opponent_signals(C_L, h_L))

    return CAM_Specification_LLAB(
        L_L,
        Ch_L,
        as_float(from_range_degrees(h_L)),
        s_L,
        C_L,
        None,
        A_L,
        B_L,
    )


def XYZ_to_RGB_LLAB(XYZ: ArrayLike) -> NDArray:
    """
    Convert from *CIE XYZ* tristimulus values to normalised cone responses.

    Parameters
    ----------
    XYZ
        *CIE XYZ* tristimulus values.

    Returns
    -------
    :class:`numpy.ndarray`
        Normalised cone responses.

    Examples
    --------
    >>> XYZ = np.array([19.01, 20.00, 21.78])
    >>> XYZ_to_RGB_LLAB(XYZ)  # doctest: +ELLIPSIS
    array([ 0.9414279...,  1.0404012...,  1.0897088...])
    """

    _X, Y, _Z = tsplit(XYZ)

    Y = tstack([Y, Y, Y])
    XYZ_n = XYZ / Y

    return vector_dot(MATRIX_XYZ_TO_RGB_LLAB, XYZ_n)


def chromatic_adaptation(
    RGB: ArrayLike,
    RGB_0: ArrayLike,
    RGB_0r: ArrayLike,
    Y: FloatingOrArrayLike,
    D: FloatingOrArrayLike = 1,
) -> NDArray:
    """
    Apply chromatic adaptation to given *RGB* normalised cone responses
    array.

    Parameters
    ----------
    RGB
        *RGB* normalised cone responses array of test sample / stimulus.
    RGB_0
        *RGB* normalised cone responses array of reference white.
    RGB_0r
        *RGB* normalised cone responses array of reference illuminant
        *CIE Standard Illuminant D Series* *D65*.
    Y
        Tristimulus values :math:`Y` of the stimulus.
    D
         *Discounting-the-Illuminant* factor normalised to domain [0, 1].

    Returns
    -------
    :class:`numpy.ndarray`
        Adapted *CIE XYZ* tristimulus values.

    Examples
    --------
    >>> RGB = np.array([0.94142795, 1.04040120, 1.08970885])
    >>> RGB_0 = np.array([0.94146023, 1.04039386, 1.08950293])
    >>> RGB_0r = np.array([0.94146023, 1.04039386, 1.08950293])
    >>> Y = 20.0
    >>> chromatic_adaptation(RGB, RGB_0, RGB_0r, Y)  # doctest: +ELLIPSIS
    array([ 19.01,  20.  ,  21.78])
    """

    R, G, B = tsplit(RGB)
    R_0, G_0, B_0 = tsplit(RGB_0)
    R_0r, G_0r, B_0r = tsplit(RGB_0r)
    Y = as_float_array(Y)
    D = as_float_array(D)

    beta = spow(B_0 / B_0r, 0.0834)

    R_r = (D * (R_0r / R_0) + 1 - D) * R
    G_r = (D * (G_0r / G_0) + 1 - D) * G
    B_r = (D * (B_0r / spow(B_0, beta)) + 1 - D) * spow(B, beta)

    RGB_r = tstack([R_r, G_r, B_r])

    Y = tstack([Y, Y, Y])

    XYZ_r = vector_dot(MATRIX_RGB_TO_XYZ_LLAB, RGB_r * Y)

    return XYZ_r


def f(x: FloatingOrArrayLike, F_S: FloatingOrArrayLike) -> FloatingOrNDArray:
    """
    Define the nonlinear response function of the *:math:`LLAB(l:c)`* colour
    appearance model used to model the nonlinear behaviour of various visual
    responses.

    Parameters
    ----------
    x
        Visual response variable :math:`x`.
    F_S
        Surround induction factor :math:`F_S`.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Modeled visual response variable :math:`x`.

    Examples
    --------
    >>> x = np.array([0.23350512, 0.23351103, 0.23355179])
    >>> f(0.200009186234000, 3)  # doctest: +ELLIPSIS
    0.5848125...
    """

    x = as_float_array(x)
    F_S = as_float_array(F_S)

    x_m = np.where(
        x > 0.008856,
        spow(x, 1 / F_S),
        ((spow(0.008856, 1 / F_S) - (16 / 116)) / 0.008856) * x + (16 / 116),
    )

    return as_float(x_m)


def opponent_colour_dimensions(
    XYZ: ArrayLike,
    Y_b: FloatingOrArrayLike,
    F_S: FloatingOrArrayLike,
    F_L: FloatingOrArrayLike,
) -> NDArray:
    """
    Return opponent colour dimensions from given adapted *CIE XYZ* tristimulus
    values.

    The opponent colour dimensions are based on a modified *CIE L\\*a\\*b\\**
    colourspace formulae.

    Parameters
    ----------
    XYZ
        Adapted *CIE XYZ* tristimulus values.
    Y_b
        Luminance factor of the background in :math:`cd/m^2`.
    F_S
        Surround induction factor :math:`F_S`.
    F_L
        Lightness induction factor :math:`F_L`.

    Returns
    -------
    :class:`numpy.ndarray`
        Opponent colour dimensions.

    Examples
    --------
    >>> XYZ = np.array([19.00999572, 20.00091862, 21.77993863])
    >>> Y_b = 20.0
    >>> F_S = 3.0
    >>> F_L = 1.0
    >>> opponent_colour_dimensions(XYZ, Y_b, F_S, F_L)  # doctest: +ELLIPSIS
    array([  3.7368047...e+01,  -4.4986443...e-03,  -5.2604647...e-03])
    """

    X, Y, Z = tsplit(XYZ)
    Y_b = as_float_array(Y_b)
    F_S = as_float_array(F_S)
    F_L = as_float_array(F_L)

    # Account for background lightness contrast.
    z = 1 + F_L * spow(Y_b / 100, 0.5)

    # Computing modified *CIE L\\*a\\*b\\** colourspace array.
    L = 116 * spow(f(Y / 100, F_S), z) - 16
    a = 500 * (f(X / 95.05, F_S) - f(Y / 100, F_S))
    b = 200 * (f(Y / 100, F_S) - f(Z / 108.88, F_S))

    Lab = tstack([L, a, b])

    return Lab


def hue_angle(
    a: FloatingOrArrayLike, b: FloatingOrArrayLike
) -> FloatingOrNDArray:
    """
    Return the *hue* angle :math:`h_L` in degrees.

    Parameters
    ----------
    a
        Opponent colour dimension :math:`a`.
    b
        Opponent colour dimension :math:`b`.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        *Hue* angle :math:`h_L` in degrees.

    Examples
    --------
    >>> hue_angle(-4.49864756e-03, -5.26046353e-03)  # doctest: +ELLIPSIS
    229.4635727...
    """

    a = as_float_array(a)
    b = as_float_array(b)

    h_L = np.degrees(np.arctan2(b, a)) % 360

    return as_float(h_L)


def chroma_correlate(
    a: FloatingOrArrayLike, b: FloatingOrArrayLike
) -> FloatingOrNDArray:
    """
    Return the correlate of *chroma* :math:`Ch_L`.

    Parameters
    ----------
    a
        Opponent colour dimension :math:`a`.
    b
        Opponent colour dimension :math:`b`.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Correlate of *chroma* :math:`Ch_L`.

    Examples
    --------
    >>> a = -4.49864756e-03
    >>> b = -5.26046353e-03
    >>> chroma_correlate(a, b)  # doctest: +ELLIPSIS
    0.0086506...
    """

    a = as_float_array(a)
    b = as_float_array(b)

    c = spow(a**2 + b**2, 0.5)
    Ch_L = 25 * np.log1p(0.05 * c)

    return as_float(Ch_L)


def colourfulness_correlate(
    L: FloatingOrArrayLike,
    L_L: FloatingOrArrayLike,
    Ch_L: FloatingOrArrayLike,
    F_C: FloatingOrArrayLike,
) -> FloatingOrNDArray:
    """
    Return the correlate of *colourfulness* :math:`C_L`.

    Parameters
    ----------
    L
        Absolute luminance :math:`L` of reference white in :math:`cd/m^2`.
    L_L
        Correlate of *Lightness* :math:`L_L`.
    Ch_L
        Correlate of *chroma* :math:`Ch_L`.
    F_C
        Chroma induction factor :math:`F_C`.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Correlate of *colourfulness* :math:`C_L`.

    Examples
    --------
    >>> L = 318.31
    >>> L_L = 37.368047493928195
    >>> Ch_L = 0.008650662051714
    >>> F_C = 1.0
    >>> colourfulness_correlate(L, L_L, Ch_L, F_C)  # doctest: +ELLIPSIS
    0.0183832...
    """

    L = as_float_array(L)
    L_L = as_float_array(L_L)
    Ch_L = as_float_array(Ch_L)
    F_C = as_float_array(F_C)

    S_C = 1 + 0.47 * np.log10(L) - 0.057 * np.log10(L) ** 2
    S_M = 0.7 + 0.02 * L_L - 0.0002 * L_L**2
    C_L = Ch_L * S_M * S_C * F_C

    return as_float(C_L)


def saturation_correlate(
    Ch_L: FloatingOrArrayLike, L_L: FloatingOrArrayLike
) -> FloatingOrNDArray:
    """
    Return the correlate of *saturation* :math:`S_L`.

    Parameters
    ----------
    Ch_L
        Correlate of *chroma* :math:`Ch_L`.
    L_L
        Correlate of *Lightness* :math:`L_L`.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Correlate of *saturation* :math:`S_L`.

    Examples
    --------
    >>> Ch_L = 0.008650662051714
    >>> L_L = 37.368047493928195
    >>> saturation_correlate(Ch_L, L_L)  # doctest: +ELLIPSIS
    0.0002314...
    """

    Ch_L = as_float_array(Ch_L)
    L_L = as_float_array(L_L)

    S_L = Ch_L / L_L

    return S_L


def final_opponent_signals(
    C_L: FloatingOrArrayLike, h_L: FloatingOrArrayLike
) -> NDArray:
    """
    Return the final opponent signals :math:`A_L` and :math:`B_L`.

    Parameters
    ----------
    C_L
        Correlate of *colourfulness* :math:`C_L`.
    h_L
        Correlate of *hue* :math:`h_L` in degrees.

    Returns
    -------
    :class:`numpy.ndarray`
        Final opponent signals :math:`A_L` and :math:`B_L`.

    Examples
    --------
    >>> C_L = 0.0183832899143
    >>> h_L = 229.46357270858391
    >>> final_opponent_signals(C_L, h_L)  # doctest: +ELLIPSIS
    array([-0.0119478..., -0.0139711...])
    """

    AB_L = polar_to_cartesian(tstack([as_float_array(C_L), np.radians(h_L)]))

    return AB_L



"""
IES TM-27-14 Data Input / Output
================================

Defines the :class:`colour.SpectralDistribution_IESTM2714` class handling *IES
TM-27-14* spectral data *XML* files.

References
----------
-   :cite:`IESComputerCommittee2014a` : IES Computer Committee, & TM-27-14
    Working Group. (2014). IES Standard Format for the Electronic Transfer of
    Spectral Data Electronic Transfer of Spectral Data. Illuminating
    Engineering Society. ISBN:978-0-87995-295-2
"""

from __future__ import annotations

import os
import re
from dataclasses import dataclass, field
from xml.etree import ElementTree  # nosec
from xml.dom import minidom  # nosec

from colour.colorimetry import SpectralDistribution
from colour.hints import (
    Any,
    ArrayLike,
    Boolean,
    Callable,
    Floating,
    Integer,
    Literal,
    Optional,
    cast,
)
from colour.utilities import (
    Structure,
    as_float_array,
    as_float_scalar,
    attest,
    optional,
    is_numeric,
    is_string,
    tstack,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "VERSION_IESTM2714",
    "NAMESPACE_IESTM2714",
    "Element_Specification_IESTM2714",
    "Header_IESTM2714",
    "SpectralDistribution_IESTM2714",
]

VERSION_IESTM2714: str = "1.0"

NAMESPACE_IESTM2714: str = "http://www.ies.org/iestm2714"


@dataclass
class Element_Specification_IESTM2714:
    """
    *IES TM-27-14* spectral data *XML* file element specification.

    Parameters
    ----------
    element
        Element name.
    attribute
        Associated attribute name.
    type_
        Element type.
    required
        Is element required.
    read_conversion
        Method to convert from *XML* to type on reading.
    write_conversion
        Method to convert from type to *XML* on writing.
    """

    element: str
    attribute: str
    type_: Any = field(default_factory=str)
    required: Boolean = field(default_factory=lambda: False)
    read_conversion: Callable = field(
        default_factory=lambda: lambda x: None if x == "None" else str(x)
    )
    write_conversion: Callable = field(default_factory=lambda: str)


class Header_IESTM2714:
    """
    Define the header object for a *IES TM-27-14* spectral distribution.

    Parameters
    ----------
    manufacturer
        Manufacturer of the device under test.
    catalog_number
        Manufacturer's product catalog number.
    description
        Description of the spectral data in the spectral data *XML* file.
    document_creator
        Creator of the spectral data *XML* file, which may be a
        test lab, a research group, a standard body, a company or an
        individual.
    unique_identifier
        Unique identifier to the product under test or the spectral data in the
        document.
    measurement_equipment
        Description of the equipment used to measure the spectral data.
    laboratory
        Testing laboratory name that performed the spectral data measurements.
    report_number
        Testing laboratory report number.
    report_date
        Testing laboratory report date using the *XML DateTime Data Type*,
        *YYYY-MM-DDThh:mm:ss*.
    document_creation_date
        Spectral data *XML* file creation date using the
        *XML DateTime Data Type*, *YYYY-MM-DDThh:mm:ss*.
    comments
        Additional information relating to the tested and reported data.

    Attributes
    ----------
    -   :attr:`~colour.io.ies_tm2714.Header_IESTM2714.mapping`
    -   :attr:`~colour.io.ies_tm2714.Header_IESTM2714.manufacturer`
    -   :attr:`~colour.io.ies_tm2714.Header_IESTM2714.catalog_number`
    -   :attr:`~colour.io.ies_tm2714.Header_IESTM2714.description`
    -   :attr:`~colour.io.ies_tm2714.Header_IESTM2714.document_creator`
    -   :attr:`~colour.io.ies_tm2714.Header_IESTM2714.unique_identifier`
    -   :attr:`~colour.io.ies_tm2714.Header_IESTM2714.measurement_equipment`
    -   :attr:`~colour.io.ies_tm2714.Header_IESTM2714.laboratory`
    -   :attr:`~colour.io.ies_tm2714.Header_IESTM2714.report_number`
    -   :attr:`~colour.io.ies_tm2714.Header_IESTM2714.report_date`
    -   :attr:`~colour.io.ies_tm2714.Header_IESTM2714.document_creation_date`
    -   :attr:`~colour.io.ies_tm2714.Header_IESTM2714.comments`

    Methods
    -------
    -   :meth:`~colour.io.ies_tm2714.Header_IESTM2714.__init__`
    -   :meth:`~colour.io.ies_tm2714.Header_IESTM2714.__hash__`
    -   :meth:`~colour.io.ies_tm2714.Header_IESTM2714.__eq__`
    -   :meth:`~colour.io.ies_tm2714.Header_IESTM2714.__ne__`

    Examples
    --------
    >>> Header_IESTM2714('colour-science')  # doctest: +ELLIPSIS
    <...Header_IESTM2714 object at 0x...>
    >>> Header_IESTM2714('colour-science').manufacturer  # doctest: +SKIP
    'colour-science'
    """

    def __init__(
        self,
        manufacturer: Optional[str] = None,
        catalog_number: Optional[str] = None,
        description: Optional[str] = None,
        document_creator: Optional[str] = None,
        unique_identifier: Optional[str] = None,
        measurement_equipment: Optional[str] = None,
        laboratory: Optional[str] = None,
        report_number: Optional[str] = None,
        report_date: Optional[str] = None,
        document_creation_date: Optional[str] = None,
        comments: Optional[str] = None,
    ):

        self._mapping: Structure = Structure(
            **{
                "element": "Header",
                "elements": (
                    Element_Specification_IESTM2714(
                        "Manufacturer", "manufacturer"
                    ),
                    Element_Specification_IESTM2714(
                        "CatalogNumber", "catalog_number"
                    ),
                    Element_Specification_IESTM2714(
                        "Description", "description", required=True
                    ),
                    Element_Specification_IESTM2714(
                        "DocumentCreator", "document_creator", required=True
                    ),
                    Element_Specification_IESTM2714(
                        "UniqueIdentifier", "unique_identifier"
                    ),
                    Element_Specification_IESTM2714(
                        "MeasurementEquipment", "measurement_equipment"
                    ),
                    Element_Specification_IESTM2714(
                        "Laboratory", "laboratory"
                    ),
                    Element_Specification_IESTM2714(
                        "ReportNumber", "report_number"
                    ),
                    Element_Specification_IESTM2714(
                        "ReportDate", "report_date"
                    ),
                    Element_Specification_IESTM2714(
                        "DocumentCreationDate",
                        "document_creation_date",
                        required=True,
                    ),
                    Element_Specification_IESTM2714(
                        "Comments", "comments", False
                    ),
                ),
            }
        )

        self._manufacturer: Optional[str] = None
        self.manufacturer = manufacturer
        self._catalog_number: Optional[str] = None
        self.catalog_number = catalog_number
        self._description: Optional[str] = None
        self.description = description
        self._document_creator: Optional[str] = None
        self.document_creator = document_creator
        self._unique_identifier: Optional[str] = None
        self.unique_identifier = unique_identifier
        self._measurement_equipment: Optional[str] = None
        self.measurement_equipment = measurement_equipment
        self._laboratory: Optional[str] = None
        self.laboratory = laboratory
        self._report_number: Optional[str] = None
        self.report_number = report_number
        self._report_date: Optional[str] = None
        self.report_date = report_date
        self._document_creation_date: Optional[str] = None
        self.document_creation_date = document_creation_date
        self._comments: Optional[str] = None
        self.comments = comments

    @property
    def mapping(self) -> Structure:
        """
        Getter property for the mapping structure.

        Returns
        -------
        :class:`colour.utilities.Structure`
            Mapping structure.
        """

        return self._mapping

    @property
    def manufacturer(self) -> Optional[str]:
        """
        Getter and setter property for the manufacturer.

        Parameters
        ----------
        value
            Value to set the manufacturer with.

        Returns
        -------
        :py:data:`None` or :class:`str`
            Manufacturer.
        """

        return self._manufacturer

    @manufacturer.setter
    def manufacturer(self, value: Optional[str]):
        """Setter for the **self.manufacturer** property."""

        if value is not None:
            attest(
                is_string(value),
                f'"manufacturer" property: "{value}" type is not "str"!',
            )

        self._manufacturer = value

    @property
    def catalog_number(self) -> Optional[str]:
        """
        Getter and setter property for the catalog number.

        Parameters
        ----------
        value
            Value to set the catalog number with.

        Returns
        -------
        :py:data:`None` or :class:`str`
            Catalog number.
        """

        return self._catalog_number

    @catalog_number.setter
    def catalog_number(self, value: Optional[str]):
        """Setter for the **self.catalog_number** property."""

        if value is not None:
            attest(
                is_string(value),
                f'"catalog_number" property: "{value}" type is not "str"!',
            )

        self._catalog_number = value

    @property
    def description(self) -> Optional[str]:
        """
        Getter and setter property for the description.

        Parameters
        ----------
        value
            Value to set the description with.

        Returns
        -------
        :py:data:`None` or :class:`str`
            Description.
        """

        return self._description

    @description.setter
    def description(self, value: Optional[str]):
        """Setter for the **self.description** property."""

        if value is not None:
            attest(
                is_string(value),
                f'"description" property: "{value}" type is not "str"!',
            )

        self._description = value

    @property
    def document_creator(self) -> Optional[str]:
        """
        Getter and setter property for the document creator.

        Parameters
        ----------
        value
            Value to set the document creator with.

        Returns
        -------
        :py:data:`None` or :class:`str`
            Document creator.
        """

        return self._document_creator

    @document_creator.setter
    def document_creator(self, value: Optional[str]):
        """Setter for the **self.document_creator** property."""

        if value is not None:
            attest(
                is_string(value),
                f'"document_creator" property: "{value}" type is not "str"!',
            )

        self._document_creator = value

    @property
    def unique_identifier(self) -> Optional[str]:
        """
        Getter and setter property for the unique identifier.

        Parameters
        ----------
        value
            Value to set the unique identifier with.

        Returns
        -------
        :py:data:`None` or :class:`str`
            Unique identifier.
        """

        return self._unique_identifier

    @unique_identifier.setter
    def unique_identifier(self, value: Optional[str]):
        """Setter for the **self.unique_identifier** property."""

        if value is not None:
            attest(
                is_string(value),
                f'"unique_identifier" property: "{value}" type is not "str"!',
            )

        self._unique_identifier = value

    @property
    def measurement_equipment(self) -> Optional[str]:
        """
        Getter and setter property for the measurement equipment.

        Parameters
        ----------
        value
            Value to set the measurement equipment with.

        Returns
        -------
        :py:data:`None` or :class:`str`
            Measurement equipment.
        """

        return self._measurement_equipment

    @measurement_equipment.setter
    def measurement_equipment(self, value: Optional[str]):
        """Setter for the **self.measurement_equipment** property."""

        if value is not None:
            attest(
                is_string(value),
                f'"measurement_equipment" property: "{value}" type is not "str"!',
            )

        self._measurement_equipment = value

    @property
    def laboratory(self) -> Optional[str]:
        """
        Getter and setter property for the laboratory.

        Parameters
        ----------
        value
            Value to set the laboratory with.

        Returns
        -------
        :py:data:`None` or :class:`str`
            Laboratory.
        """

        return self._laboratory

    @laboratory.setter
    def laboratory(self, value: Optional[str]):
        """Setter for the **self.measurement_equipment** property."""

        if value is not None:
            attest(
                is_string(value),
                f'"laboratory" property: "{value}" type is not "str"!',
            )

        self._laboratory = value

    @property
    def report_number(self) -> Optional[str]:
        """
        Getter and setter property for the report number.

        Parameters
        ----------
        value
            Value to set the report number with.

        Returns
        -------
        :py:data:`None` or :class:`str`
            Report number.
        """

        return self._report_number

    @report_number.setter
    def report_number(self, value: Optional[str]):
        """Setter for the **self.report_number** property."""

        if value is not None:
            attest(
                is_string(value),
                f'"report_number" property: "{value}" type is not "str"!',
            )

        self._report_number = value

    @property
    def report_date(self) -> Optional[str]:
        """
        Getter and setter property for the report date.

        Parameters
        ----------
        value
            Value to set the report date with.

        Returns
        -------
        :py:data:`None` or :class:`str`
            Report date.
        """

        return self._report_date

    @report_date.setter
    def report_date(self, value: Optional[str]):
        """Setter for the **self.report_date** property."""

        if value is not None:
            attest(
                is_string(value),
                f'"report_date" property: "{value}" type is not "str"!',
            )

        self._report_date = value

    @property
    def document_creation_date(self) -> Optional[str]:
        """
        Getter and setter property for the document creation date.

        Parameters
        ----------
        value
            Value to set the document creation date with.

        Returns
        -------
        :py:data:`None` or :class:`str`
            Document creation date.
        """

        return self._document_creation_date

    @document_creation_date.setter
    def document_creation_date(self, value: Optional[str]):
        """Setter for the **self.document_creation_date** property."""

        if value is not None:
            attest(
                is_string(value),
                f'"document_creation_date" property: "{value}" type is not "str"!',
            )

        self._document_creation_date = value

    @property
    def comments(self) -> Optional[str]:
        """
        Getter and setter property for the comments.

        Parameters
        ----------
        value
            Value to set the comments with.

        Returns
        -------
        :py:data:`None` or :class:`str`
            Comments.
        """

        return self._comments

    @comments.setter
    def comments(self, value: Optional[str]):
        """Setter for the **self.comments** property."""

        if value is not None:
            attest(
                is_string(value),
                f'"comments" property: "{value}" type is not "str"!',
            )

        self._comments = value

    def __hash__(self) -> Integer:
        """
        Return the header hash.

        Returns
        -------
        :class:`numpy.integer`
            Object hash.
        """

        return hash(
            (
                self._manufacturer,
                self._catalog_number,
                self._description,
                self._document_creator,
                self._unique_identifier,
                self._measurement_equipment,
                self._laboratory,
                self._report_number,
                self._report_date,
                self._document_creation_date,
                self._comments,
            )
        )

    def __eq__(self, other: Any) -> Boolean:
        """
        Return whether the header is equal to given other object.

        Parameters
        ----------
        other
            Object to test whether it is equal to the header.

        Returns
        -------
        :class:`bool`
            Whether given object is equal to the header.

        Examples
        --------
        >>> Header_IESTM2714('Foo') == Header_IESTM2714('Foo')
        True
        >>> Header_IESTM2714('Foo') == Header_IESTM2714('Bar')
        False
        """

        if isinstance(other, Header_IESTM2714):
            return all(
                [
                    self._manufacturer == other.manufacturer,
                    self._catalog_number == other.catalog_number,
                    self._description == other.description,
                    self._document_creator == other.document_creator,
                    self._unique_identifier == other.unique_identifier,
                    self._measurement_equipment == other.measurement_equipment,
                    self._laboratory == other.laboratory,
                    self._report_number == other.report_number,
                    self._report_date == other.report_date,
                    self._document_creation_date
                    == other.document_creation_date,
                    self._comments == other.comments,
                ]
            )
        return False

    def __ne__(self, other: Any) -> Boolean:
        """
        Return whether the header is not equal to given other object.

        Parameters
        ----------
        other
            Object to test whether it is not equal to the header.

        Returns
        -------
        :class:`bool`
            Whether given object is not equal to the header.

        Examples
        --------
        >>> Header_IESTM2714('Foo') != Header_IESTM2714('Foo')
        False
        >>> Header_IESTM2714('Foo') != Header_IESTM2714('Bar')
        True
        """

        return not (self == other)


class SpectralDistribution_IESTM2714(SpectralDistribution):
    """
    Define a *IES TM-27-14* spectral distribution.

    This class can read and write *IES TM-27-14* spectral data *XML* files.

    Parameters
    ----------
    path
        Spectral data *XML* file path.
    header
        *IES TM-27-14* spectral distribution header.
    spectral_quantity
        Quantity of measurement for each element of the spectral data.
    reflection_geometry
        Spectral reflectance factors geometric conditions.
    transmission_geometry
        Spectral transmittance factors geometric conditions.
    bandwidth_FWHM
        Spectroradiometer full-width half-maximum bandwidth in nanometers.
    bandwidth_corrected
        Specifies if bandwidth correction has been applied to the measured
        data.

    Other Parameters
    ----------------
    data
        Data to be stored in the spectral distribution.
    domain
        Values to initialise the
        :attr:`colour.SpectralDistribution.wavelength` property with.
        If both ``data`` and ``domain`` arguments are defined, the latter will
        be used to initialise the
        :attr:`colour.SpectralDistribution.wavelength` property.
    extrapolator
        Extrapolator class type to use as extrapolating function.
    extrapolator_kwargs
        Arguments to use when instantiating the extrapolating function.
    interpolator
        Interpolator class type to use as interpolating function.
    interpolator_kwargs
        Arguments to use when instantiating the interpolating function.
    name
        Spectral distribution name.
    strict_name
        Spectral distribution name for figures, default to
        :attr:`colour.SpectralDistribution.name` property value.

    Notes
    -----
    *Reflection Geometry*

    -   di:8: Diffuse / eight-degree, specular component included.
    -   de:8: Diffuse / eight-degree, specular component excluded.
    -   8:di: Eight-degree / diffuse, specular component included.
    -   8:de: Eight-degree / diffuse, specular component excluded.
    -   d:d: Diffuse / diffuse.
    -   d:0: Alternative diffuse.
    -   45a:0: Forty-five degree annular / normal.
    -   45c:0: Forty-five degree circumferential / normal.
    -   0:45a: Normal / forty-five degree annular.
    -   45x:0: Forty-five degree directional / normal.
    -   0:45x: Normal / forty-five degree directional.
    -   other: User-specified in comments.

    *Transmission Geometry*

    -   0:0: Normal / normal.
    -   di:0: Diffuse / normal, regular component included.
    -   de:0: Diffuse / normal, regular component excluded.
    -   0:di: Normal / diffuse, regular component included.
    -   0:de: Normal / diffuse, regular component excluded.
    -   d:d: Diffuse / diffuse.
    -   other: User-specified in comments.

    Attributes
    ----------
    -   :attr:`~colour.SpectralDistribution_IESTM2714.mapping`
    -   :attr:`~colour.SpectralDistribution_IESTM2714.path`
    -   :attr:`~colour.SpectralDistribution_IESTM2714.header`
    -   :attr:`~colour.SpectralDistribution_IESTM2714.spectral_quantity`
    -   :attr:`~colour.SpectralDistribution_IESTM2714.reflection_geometry`
    -   :attr:`~colour.SpectralDistribution_IESTM2714.transmission_geometry`
    -   :attr:`~colour.SpectralDistribution_IESTM2714.bandwidth_FWHM`
    -   :attr:`~colour.SpectralDistribution_IESTM2714.bandwidth_corrected`

    Methods
    -------
    -   :meth:`~colour.SpectralDistribution_IESTM2714.__init__`
    -   :meth:`~colour.SpectralDistribution_IESTM2714.read`
    -   :meth:`~colour.SpectralDistribution_IESTM2714.write`

    References
    ----------
    :cite:`IESComputerCommittee2014a`

    Examples
    --------
    >>> from os.path import dirname, join
    >>> directory = join(dirname(__file__), 'tests', 'resources')
    >>> sd = SpectralDistribution_IESTM2714(
    ...     join(directory, 'Fluorescent.spdx')).read()
    >>> sd.name  # doctest: +SKIP
    'Unknown - N/A - Rare earth fluorescent lamp'
    >>> sd.header.comments
    'Ambient temperature 25 degrees C.'
    >>> sd[501.7]  # doctest: +ELLIPSIS
    0.0950000...
    """

    def __init__(
        self,
        path: Optional[str] = None,
        header: Optional[Header_IESTM2714] = None,
        spectral_quantity: Optional[
            Literal[
                "absorptance",
                "exitance",
                "flux",
                "intensity",
                "irradiance",
                "radiance",
                "reflectance",
                "relative",
                "transmittance",
                "R-Factor",
                "T-Factor",
                "other",
            ]
        ] = None,
        reflection_geometry: Optional[
            Literal[
                "di:8",
                "de:8",
                "8:di",
                "8:de",
                "d:d",
                "d:0",
                "45a:0",
                "45c:0",
                "0:45a",
                "45x:0",
                "0:45x",
                "other",
            ]
        ] = None,
        transmission_geometry: Optional[
            Literal["0:0", "di:0", "de:0", "0:di", "0:de", "d:d", "other"]
        ] = None,
        bandwidth_FWHM: Optional[Floating] = None,
        bandwidth_corrected: Optional[Boolean] = None,
        **kwargs,
    ):

        super().__init__(**kwargs)

        self._mapping: Structure = Structure(
            **{
                "element": "SpectralDistribution",
                "elements": (
                    Element_Specification_IESTM2714(
                        "SpectralQuantity", "spectral_quantity", required=True
                    ),
                    Element_Specification_IESTM2714(
                        "ReflectionGeometry", "reflection_geometry"
                    ),
                    Element_Specification_IESTM2714(
                        "TransmissionGeometry", "transmission_geometry"
                    ),
                    Element_Specification_IESTM2714(
                        "BandwidthFWHM",
                        "bandwidth_FWHM",
                        read_conversion=as_float_scalar,
                    ),
                    Element_Specification_IESTM2714(
                        "BandwidthCorrected",
                        "bandwidth_corrected",
                        read_conversion=(
                            lambda x: True if x == "true" else False
                        ),
                        write_conversion=(
                            lambda x: "true" if x is True else "False"
                        ),
                    ),
                ),
                "data": Element_Specification_IESTM2714(
                    "SpectralData", "wavelength", required=True
                ),
            }
        )

        self._path: Optional[str] = None
        self.path = path
        self._header: Header_IESTM2714 = Header_IESTM2714()
        self.header = optional(header, self._header)
        self._spectral_quantity: Optional[
            Literal[
                "absorptance",
                "exitance",
                "flux",
                "intensity",
                "irradiance",
                "radiance",
                "reflectance",
                "relative",
                "transmittance",
                "R-Factor",
                "T-Factor",
                "other",
            ]
        ] = None
        self.spectral_quantity = spectral_quantity
        self._reflection_geometry: Optional[
            Literal[
                "di:8",
                "de:8",
                "8:di",
                "8:de",
                "d:d",
                "d:0",
                "45a:0",
                "45c:0",
                "0:45a",
                "45x:0",
                "0:45x",
                "other",
            ]
        ] = None
        self.reflection_geometry = reflection_geometry
        self._transmission_geometry: Optional[
            Literal["0:0", "di:0", "de:0", "0:di", "0:de", "d:d", "other"]
        ] = None
        self.transmission_geometry = transmission_geometry
        self._bandwidth_FWHM: Optional[Floating] = None
        self.bandwidth_FWHM = bandwidth_FWHM
        self._bandwidth_corrected: Optional[Boolean] = None
        self.bandwidth_corrected = bandwidth_corrected

    @property
    def mapping(self) -> Structure:
        """
        Getter property for the mapping structure.

        Returns
        -------
        :class:`colour.utilities.Structure`
            Mapping structure.
        """

        return self._mapping

    @property
    def path(self) -> Optional[str]:
        """
        Getter and setter property for the path.

        Parameters
        ----------
        value
            Value to set the path with.

        Returns
        -------
        :py:data:`None` or :class:`str`
            Path.
        """

        return self._path

    @path.setter
    def path(self, value: Optional[str]):
        """Setter for the **self.path** property."""

        if value is not None:
            attest(
                is_string(value),
                f'"path" property: "{value}" type is not "str"!',
            )

        self._path = value

    @property
    def header(self) -> Header_IESTM2714:
        """
        Getter and setter property for the header.

        Parameters
        ----------
        value
            Value to set the header with.

        Returns
        -------
        :class:`colour.io.tm2714.Header_IESTM2714`
            Header.
        """

        return self._header

    @header.setter
    def header(self, value: Header_IESTM2714):
        """Setter for the **self.header** property."""

        attest(
            isinstance(value, Header_IESTM2714),
            f'"header" property: "{value}" type is not "Header_IESTM2714"!',
        )

        self._header = value

    @property
    def spectral_quantity(
        self,
    ) -> Optional[
        Literal[
            "absorptance",
            "exitance",
            "flux",
            "intensity",
            "irradiance",
            "radiance",
            "reflectance",
            "relative",
            "transmittance",
            "R-Factor",
            "T-Factor",
            "other",
        ]
    ]:
        """
        Getter and setter property for the spectral quantity.

        Parameters
        ----------
        value
            Value to set the spectral quantity with.

        Returns
        -------
        :py:data:`None` or :class:`str`
            Spectral quantity.
        """

        return self._spectral_quantity

    @spectral_quantity.setter
    def spectral_quantity(
        self,
        value: Optional[
            Literal[
                "absorptance",
                "exitance",
                "flux",
                "intensity",
                "irradiance",
                "radiance",
                "reflectance",
                "relative",
                "transmittance",
                "R-Factor",
                "T-Factor",
                "other",
            ]
        ],
    ):
        """Setter for the **self.spectral_quantity** property."""

        if value is not None:
            attest(
                is_string(value),
                f'"spectral_quantity" property: "{value}" type is not "str"!',
            )

        self._spectral_quantity = value

    @property
    def reflection_geometry(
        self,
    ) -> Optional[
        Literal[
            "di:8",
            "de:8",
            "8:di",
            "8:de",
            "d:d",
            "d:0",
            "45a:0",
            "45c:0",
            "0:45a",
            "45x:0",
            "0:45x",
            "other",
        ]
    ]:
        """
        Getter and setter property for the reflection geometry.

        Parameters
        ----------
        value
            Value to set the reflection geometry with.

        Returns
        -------
        :py:data:`None` or :class:`str`
            Reflection geometry.
        """

        return self._reflection_geometry

    @reflection_geometry.setter
    def reflection_geometry(
        self,
        value: Optional[
            Literal[
                "di:8",
                "de:8",
                "8:di",
                "8:de",
                "d:d",
                "d:0",
                "45a:0",
                "45c:0",
                "0:45a",
                "45x:0",
                "0:45x",
                "other",
            ]
        ],
    ):
        """Setter for the **self.reflection_geometry** property."""

        if value is not None:
            attest(
                is_string(value),
                f'"reflection_geometry" property: "{value}" type is not "str"!',
            )

        self._reflection_geometry = value

    @property
    def transmission_geometry(
        self,
    ) -> Optional[
        Literal["0:0", "di:0", "de:0", "0:di", "0:de", "d:d", "other"]
    ]:
        """
        Getter and setter property for the transmission geometry.

        Parameters
        ----------
        value
            Value to set the transmission geometry with.

        Returns
        -------
        :py:data:`None` or :class:`str`
            Transmission geometry.
        """

        return self._transmission_geometry

    @transmission_geometry.setter
    def transmission_geometry(
        self,
        value: Optional[
            Literal["0:0", "di:0", "de:0", "0:di", "0:de", "d:d", "other"]
        ],
    ):
        """Setter for the **self.transmission_geometry** property."""

        if value is not None:
            attest(
                is_string(value),
                f'"transmission_geometry" property: "{value}" type is not "str"!',
            )

        self._transmission_geometry = value

    @property
    def bandwidth_FWHM(self) -> Optional[Floating]:
        """
        Getter and setter property for the full-width half-maximum bandwidth.

        Parameters
        ----------
        value
            Value to set the full-width half-maximum bandwidth with.

        Returns
        -------
        :py:data:`None` or :class:`numpy.floating`
            Full-width half-maximum bandwidth.
        """

        return self._bandwidth_FWHM

    @bandwidth_FWHM.setter
    def bandwidth_FWHM(self, value: Optional[Floating]):
        """Setter for the **self.bandwidth_FWHM** property."""

        if value is not None:
            attest(
                is_numeric(value),
                f'"bandwidth_FWHM" property: "{value}" is not a "number"!',
            )

            value = as_float_scalar(value)

        self._bandwidth_FWHM = value

    @property
    def bandwidth_corrected(self) -> Optional[Boolean]:
        """
        Getter and setter property for whether bandwidth correction has been
        applied to the measured data.

        Parameters
        ----------
        value
            Whether bandwidth correction has been applied to the measured data.

        Returns
        -------
        :py:data:`None` or :class:`bool`
            Whether bandwidth correction has been applied to the measured data.
        """

        return self._bandwidth_corrected

    @bandwidth_corrected.setter
    def bandwidth_corrected(self, value: Optional[Boolean]):
        """Setter for the **self.bandwidth_corrected** property."""

        if value is not None:
            attest(
                isinstance(value, bool),
                f'"bandwidth_corrected" property: "{value}" type is not "bool"!',
            )

        self._bandwidth_corrected = value

    def read(self) -> SpectralDistribution_IESTM2714:
        """
        Read and parses the spectral data *XML* file path.

        Returns
        -------
        :class:`colour.SpectralDistribution_IESTM2714`
            *IES TM-27-14* spectral distribution.

        Raises
        ------
        ValueError
            If the *IES TM-27-14* spectral distribution path is undefined.

        Examples
        --------
        >>> from os.path import dirname, join
        >>> directory = join(dirname(__file__), 'tests', 'resources')
        >>> sd = SpectralDistribution_IESTM2714(
        ...     join(directory, 'Fluorescent.spdx')).read()
        >>> sd.name  # doctest: +SKIP
        'Unknown - N/A - Rare earth fluorescent lamp'
        >>> sd.header.comments
        'Ambient temperature 25 degrees C.'
        >>> sd[400]  # doctest: +ELLIPSIS
        0.0340000...
        """

        if self._path is not None:
            formatter = "./{{{0}}}{1}/{{{0}}}{2}"

            tree = ElementTree.parse(self._path)  # nosec
            root = tree.getroot()

            match = re.match("{(.*)}", root.tag)
            if match:
                namespace = match.group(1)
            else:
                raise ValueError(
                    'The "IES TM-27-14" spectral distribution namespace '
                    "was not found!"
                )

            self.name = os.path.splitext(os.path.basename(self._path))[0]

            iterator = root.iter

            for header_element in (self.header, self):
                mapping = header_element.mapping  # type: ignore[attr-defined]
                for specification in mapping.elements:
                    element = root.find(
                        formatter.format(
                            namespace, mapping.element, specification.element
                        )
                    )
                    if element is not None:
                        setattr(
                            header_element,
                            specification.attribute,
                            specification.read_conversion(element.text),
                        )

            # Reading spectral data.
            wavelengths = []
            values = []
            for spectral_data in iterator(
                f"{{{namespace}}}{self.mapping.data.element}"
            ):
                wavelengths.append(
                    spectral_data.attrib[self.mapping.data.attribute]
                )
                values.append(spectral_data.text)

            components = [
                component
                for component in (
                    self.header.manufacturer,
                    self.header.catalog_number,
                    self.header.description,
                )
                if component is not None
            ]
            self.name = (
                "Undefined" if len(components) == 0 else " - ".join(components)
            )

            self.wavelengths = as_float_array(wavelengths)
            self.values = as_float_array(cast(ArrayLike, values))

            return self
        else:
            raise ValueError(
                'The "IES TM-27-14" spectral distribution path is undefined!'
            )

    def write(self) -> Boolean:
        """
        Write the spectral distribution spectral data to *XML* file path.

        Returns
        -------
        :class:`bool`
            Definition success.

        Examples
        --------
        >>> from os.path import dirname, join
        >>> from shutil import rmtree
        >>> from tempfile import mkdtemp
        >>> directory = join(dirname(__file__), 'tests', 'resources')
        >>> sd = SpectralDistribution_IESTM2714(
        ...     join(directory, 'Fluorescent.spdx')).read()
        >>> temporary_directory = mkdtemp()
        >>> sd.path = join(temporary_directory, 'Fluorescent.spdx')
        >>> sd.write()
        True
        >>> rmtree(temporary_directory)
        """

        if self._path is not None:
            root = ElementTree.Element("IESTM2714")
            root.attrib = {
                "xmlns": NAMESPACE_IESTM2714,
                "version": VERSION_IESTM2714,
            }

            spectral_distribution = ElementTree.Element("")
            for header_element in (self.header, self):
                mapping = header_element.mapping  # type: ignore[attr-defined]
                element = ElementTree.SubElement(root, mapping.element)
                for specification in mapping.elements:
                    element_child = ElementTree.SubElement(
                        element, specification.element
                    )
                    value = getattr(header_element, specification.attribute)
                    element_child.text = specification.write_conversion(value)

                if header_element is self:
                    spectral_distribution = element

            # Writing spectral data.
            for (wavelength, value) in tstack([self.wavelengths, self.values]):
                element_child = ElementTree.SubElement(
                    spectral_distribution, mapping.data.element
                )
                element_child.text = mapping.data.write_conversion(value)
                element_child.attrib = {
                    mapping.data.attribute: mapping.data.write_conversion(
                        wavelength
                    )
                }

            xml = minidom.parseString(
                ElementTree.tostring(root)
            ).toprettyxml()  # nosec

            with open(self._path, "w") as file:
                file.write(xml)

            return True
        else:
            raise ValueError(
                'The "IES TM-27-14" spectral distribution path is undefined!'
            )



from __future__ import annotations

from colour.utilities import CaseInsensitiveMapping
from .aces import (
    RGB_COLOURSPACE_ACES2065_1,
    RGB_COLOURSPACE_ACESCC,
    RGB_COLOURSPACE_ACESCCT,
    RGB_COLOURSPACE_ACESCG,
    RGB_COLOURSPACE_ACESPROXY,
)
from .adobe_rgb_1998 import RGB_COLOURSPACE_ADOBE_RGB1998
from .adobe_wide_gamut_rgb import RGB_COLOURSPACE_ADOBE_WIDE_GAMUT_RGB
from .apple_rgb import RGB_COLOURSPACE_APPLE_RGB
from .arri_alexa_wide_gamut import RGB_COLOURSPACE_ALEXA_WIDE_GAMUT
from .best_rgb import RGB_COLOURSPACE_BEST_RGB
from .beta_rgb import RGB_COLOURSPACE_BETA_RGB
from .blackmagic_design import RGB_COLOURSPACE_BLACKMAGIC_WIDE_GAMUT
from .cie_rgb import RGB_COLOURSPACE_CIE_RGB
from .canon_cinema_gamut import RGB_COLOURSPACE_CINEMA_GAMUT
from .color_match_rgb import RGB_COLOURSPACE_COLOR_MATCH_RGB
from .davinci_wide_gamut import RGB_COLOURSPACE_DAVINCI_WIDE_GAMUT
from .dcdm_xyz import RGB_COLOURSPACE_DCDM_XYZ
from .dci_p3 import RGB_COLOURSPACE_DCI_P3, RGB_COLOURSPACE_DCI_P3_P
from .display_p3 import RGB_COLOURSPACE_DISPLAY_P3
from .p3_d65 import RGB_COLOURSPACE_P3_D65
from .don_rgb_4 import RGB_COLOURSPACE_DON_RGB_4
from .dji_dgamut import RGB_COLOURSPACE_DJI_D_GAMUT
from .eci_rgb_v2 import RGB_COLOURSPACE_ECI_RGB_V2
from .ekta_space_ps5 import RGB_COLOURSPACE_EKTA_SPACE_PS_5
from .fujifilm_f_gamut import RGB_COLOURSPACE_F_GAMUT
from .filmlight_egamut import RGB_COLOURSPACE_FILMLIGHT_E_GAMUT
from .gopro import RGB_COLOURSPACE_PROTUNE_NATIVE
from .itur_bt_470 import RGB_COLOURSPACE_BT470_525, RGB_COLOURSPACE_BT470_625
from .itur_bt_709 import RGB_COLOURSPACE_BT709
from .itur_bt_2020 import RGB_COLOURSPACE_BT2020
from .max_rgb import RGB_COLOURSPACE_MAX_RGB
from .nikon_n_gamut import RGB_COLOURSPACE_N_GAMUT
from .pal_secam import RGB_COLOURSPACE_PAL_SECAM
from .red import (
    RGB_COLOURSPACE_RED_COLOR,
    RGB_COLOURSPACE_RED_COLOR_2,
    RGB_COLOURSPACE_RED_COLOR_3,
    RGB_COLOURSPACE_RED_COLOR_4,
    RGB_COLOURSPACE_DRAGON_COLOR,
    RGB_COLOURSPACE_DRAGON_COLOR_2,
    RGB_COLOURSPACE_RED_WIDE_GAMUT_RGB,
)
from .rimm_romm_rgb import (
    RGB_COLOURSPACE_ROMM_RGB,
    RGB_COLOURSPACE_RIMM_RGB,
    RGB_COLOURSPACE_ERIMM_RGB,
    RGB_COLOURSPACE_PROPHOTO_RGB,
)
from .russell_rgb import RGB_COLOURSPACE_RUSSELL_RGB
from .sharp import RGB_COLOURSPACE_SHARP_RGB
from .smpte_240m import RGB_COLOURSPACE_SMPTE_240M
from .smpte_c import RGB_COLOURSPACE_SMPTE_C
from .ntsc import RGB_COLOURSPACE_NTSC1953, RGB_COLOURSPACE_NTSC1987
from .sony import (
    RGB_COLOURSPACE_S_GAMUT,
    RGB_COLOURSPACE_S_GAMUT3,
    RGB_COLOURSPACE_S_GAMUT3_CINE,
    RGB_COLOURSPACE_VENICE_S_GAMUT3,
    RGB_COLOURSPACE_VENICE_S_GAMUT3_CINE,
)
from .srgb import RGB_COLOURSPACE_sRGB
from .panasonic_v_gamut import RGB_COLOURSPACE_V_GAMUT
from .xtreme_rgb import RGB_COLOURSPACE_XTREME_RGB

from colour.models.rgb import RGB_Colourspace

RGB_COLOURSPACES: CaseInsensitiveMapping = CaseInsensitiveMapping(
    dict(
        sorted(
            (colourspace.name, colourspace)
            for colourspace in locals().values()
            if isinstance(colourspace, RGB_Colourspace)
        )
    )
)
RGB_COLOURSPACES.__doc__ = """
Aggregated *RGB* colourspaces.

Aliases:

-   'aces': RGB_COLOURSPACE_ACES2065_1.name
-   'adobe1998': RGB_COLOURSPACE_ADOBE_RGB1998.name
-   'prophoto': RGB_COLOURSPACE_PROPHOTO_RGB.name
"""

RGB_COLOURSPACES["aces"] = RGB_COLOURSPACES[RGB_COLOURSPACE_ACES2065_1.name]
RGB_COLOURSPACES["adobe1998"] = RGB_COLOURSPACES[
    RGB_COLOURSPACE_ADOBE_RGB1998.name
]
RGB_COLOURSPACES["prophoto"] = RGB_COLOURSPACES[
    RGB_COLOURSPACE_PROPHOTO_RGB.name
]
# yapf: enable

__all__ = [
    "RGB_COLOURSPACES",
]
__all__ += [
    "RGB_COLOURSPACE_ACES2065_1",
    "RGB_COLOURSPACE_ACESCC",
    "RGB_COLOURSPACE_ACESCCT",
    "RGB_COLOURSPACE_ACESPROXY",
    "RGB_COLOURSPACE_ACESCG",
    "RGB_COLOURSPACE_ADOBE_RGB1998",
    "RGB_COLOURSPACE_ADOBE_WIDE_GAMUT_RGB",
    "RGB_COLOURSPACE_ALEXA_WIDE_GAMUT",
    "RGB_COLOURSPACE_APPLE_RGB",
    "RGB_COLOURSPACE_BEST_RGB",
    "RGB_COLOURSPACE_BETA_RGB",
    "RGB_COLOURSPACE_BLACKMAGIC_WIDE_GAMUT",
    "RGB_COLOURSPACE_BT470_525",
    "RGB_COLOURSPACE_BT470_625",
    "RGB_COLOURSPACE_BT709",
    "RGB_COLOURSPACE_BT2020",
    "RGB_COLOURSPACE_CIE_RGB",
    "RGB_COLOURSPACE_CINEMA_GAMUT",
    "RGB_COLOURSPACE_COLOR_MATCH_RGB",
    "RGB_COLOURSPACE_DAVINCI_WIDE_GAMUT",
    "RGB_COLOURSPACE_DCDM_XYZ",
    "RGB_COLOURSPACE_DCI_P3",
    "RGB_COLOURSPACE_DCI_P3_P",
    "RGB_COLOURSPACE_DISPLAY_P3",
    "RGB_COLOURSPACE_DJI_D_GAMUT",
    "RGB_COLOURSPACE_DON_RGB_4",
    "RGB_COLOURSPACE_ECI_RGB_V2",
    "RGB_COLOURSPACE_EKTA_SPACE_PS_5",
    "RGB_COLOURSPACE_FILMLIGHT_E_GAMUT",
    "RGB_COLOURSPACE_PROTUNE_NATIVE",
    "RGB_COLOURSPACE_MAX_RGB",
    "RGB_COLOURSPACE_N_GAMUT",
    "RGB_COLOURSPACE_P3_D65",
    "RGB_COLOURSPACE_PAL_SECAM",
    "RGB_COLOURSPACE_RED_COLOR",
    "RGB_COLOURSPACE_RED_COLOR_2",
    "RGB_COLOURSPACE_RED_COLOR_3",
    "RGB_COLOURSPACE_RED_COLOR_4",
    "RGB_COLOURSPACE_DRAGON_COLOR",
    "RGB_COLOURSPACE_DRAGON_COLOR_2",
    "RGB_COLOURSPACE_RED_WIDE_GAMUT_RGB",
    "RGB_COLOURSPACE_ROMM_RGB",
    "RGB_COLOURSPACE_RIMM_RGB",
    "RGB_COLOURSPACE_ERIMM_RGB",
    "RGB_COLOURSPACE_PROPHOTO_RGB",
    "RGB_COLOURSPACE_RUSSELL_RGB",
    "RGB_COLOURSPACE_SHARP_RGB",
    "RGB_COLOURSPACE_SMPTE_240M",
    "RGB_COLOURSPACE_SMPTE_C",
    "RGB_COLOURSPACE_NTSC1953",
    "RGB_COLOURSPACE_NTSC1987",
    "RGB_COLOURSPACE_S_GAMUT",
    "RGB_COLOURSPACE_S_GAMUT3",
    "RGB_COLOURSPACE_S_GAMUT3_CINE",
    "RGB_COLOURSPACE_VENICE_S_GAMUT3",
    "RGB_COLOURSPACE_VENICE_S_GAMUT3_CINE",
    "RGB_COLOURSPACE_sRGB",
    "RGB_COLOURSPACE_V_GAMUT",
    "RGB_COLOURSPACE_XTREME_RGB",
    "RGB_COLOURSPACE_F_GAMUT",
]



"""
ANSI/IES TM-30-18 Colour Rendition Report
=========================================

Defines the *ANSI/IES TM-30-18 Colour Rendition Report* plotting objects:

-   :func:`colour.plotting.tm3018.plot_single_sd_colour_rendition_report_full`
-   :func:`colour.plotting.
tm3018.plot_single_sd_colour_rendition_report_intermediate`
-   :func:`colour.plotting.
tm3018.plot_single_sd_colour_rendition_report_simple`
-   :func:`colour.plotting.plot_single_sd_colour_rendition_report`
"""

from __future__ import annotations

import matplotlib.pyplot as plt

from colour.colorimetry import SpectralDistribution, sd_to_XYZ
from colour.hints import Any, Dict, Literal, Optional, Tuple, Union, cast
from colour.io import SpectralDistribution_IESTM2714
from colour.models import XYZ_to_xy, XYZ_to_Luv, Luv_to_uv
from colour.plotting.tm3018.components import (
    plot_spectra_ANSIIESTM3018,
    plot_colour_vector_graphic,
    plot_local_chroma_shifts,
    plot_local_hue_shifts,
    plot_local_colour_fidelities,
    plot_colour_fidelity_indexes,
)
from colour.quality import (
    ColourQuality_Specification_ANSIIESTM3018,
    ColourRendering_Specification_CRI,
    colour_fidelity_index_ANSIIESTM3018,
    colour_rendering_index,
)
from colour.plotting import CONSTANTS_COLOUR_STYLE, override_style, render
from colour.utilities import describe_environment, optional, validate_method

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "CONSTANT_REPORT_SIZE_FULL",
    "CONSTANT_REPORT_ROW_HEIGHT_RATIOS_FULL",
    "CONSTANT_REPORT_PADDING_FULL",
    "CONSTANT_REPORT_SIZE_INTERMEDIATE",
    "CONSTANT_REPORT_ROW_HEIGHT_RATIOS_INTERMEDIATE",
    "CONSTANT_REPORT_PADDING_INTERMEDIATE",
    "CONSTANT_REPORT_SIZE_SIMPLE",
    "CONSTANT_REPORT_ROW_HEIGHT_RATIOS_SIMPLE",
    "CONSTANT_REPORT_PADDING_SIMPLE",
    "CONSTANTS_REPORT_STYLE",
    "CONTENT_REPORT_HEADER",
    "CONTENT_REPORT_FOOTER",
    "plot_single_sd_colour_rendition_report_full",
    "plot_single_sd_colour_rendition_report_intermediate",
    "plot_single_sd_colour_rendition_report_simple",
    "plot_single_sd_colour_rendition_report",
]

# Full Report Size Constants
CONSTANT_REPORT_SIZE_FULL: Tuple = (8.27, 11.69)
"""Full report size, default to A4 paper size in inches."""

CONSTANT_REPORT_ROW_HEIGHT_RATIOS_FULL: Tuple = (1, 2, 24, 3, 1)
"""Full report size row height ratios."""

CONSTANT_REPORT_PADDING_FULL: Dict = {
    "w_pad": 20 / 100,
    "h_pad": 10 / 100,
    "hspace": 0,
    "wspace": 0,
}
"""
Full report box padding, tries to define the padding around the figure and
in-between the axes.
"""

# Intermediate Report Size Constants
CONSTANT_REPORT_SIZE_INTERMEDIATE: Tuple = (8.27, 11.69 / 2.35)
"""Intermediate report size, a window into A4 paper size in inches."""

CONSTANT_REPORT_ROW_HEIGHT_RATIOS_INTERMEDIATE: Tuple = (1, 8, 1)
"""Intermediate report size row height ratios."""

CONSTANT_REPORT_PADDING_INTERMEDIATE: Dict = {
    "w_pad": 20 / 100,
    "h_pad": 10 / 100,
    "hspace": 0,
    "wspace": 0,
}
"""
Intermediate report box padding, tries to define the padding around the figure
and in-between the axes.
"""

# Simple Report Size Constants
CONSTANT_REPORT_SIZE_SIMPLE: Tuple = (8.27, 8.27)
"""Simple report size, a window into A4 paper size in inches."""

CONSTANT_REPORT_ROW_HEIGHT_RATIOS_SIMPLE: Tuple = (1, 8, 1)
"""Simple report size row height ratios."""

CONSTANT_REPORT_PADDING_SIMPLE: Dict = {
    "w_pad": 20 / 100,
    "h_pad": 10 / 100,
    "hspace": 0,
    "wspace": 0,
}
"""
Simple report box padding, tries to define the padding around the figure
and in-between the axes.
"""

CONSTANTS_REPORT_STYLE: Dict = {
    "axes.grid": False,
    "axes.labelpad": CONSTANTS_COLOUR_STYLE.geometry.short * 3,
    "axes.labelsize": "x-small",
    "axes.labelweight": "bold",
    "legend.frameon": False,
    "xtick.labelsize": "x-small",
    "ytick.labelsize": "x-small",
    "xtick.direction": "out",
    "ytick.direction": "out",
    "xtick.major.size": CONSTANTS_COLOUR_STYLE.geometry.long * 0.5,
    "ytick.major.size": CONSTANTS_COLOUR_STYLE.geometry.long * 0.5,
    "xtick.minor.size": CONSTANTS_COLOUR_STYLE.geometry.long * 0.25,
    "ytick.minor.size": CONSTANTS_COLOUR_STYLE.geometry.long * 0.25,
    "xtick.minor.visible": False,
    "ytick.minor.visible": False,
}
"""Report style overrides."""

CONTENT_REPORT_HEADER: str = "IES TM-30-18 Colour Rendition Report"
"""Report header content, i.e. the report title."""

CONTENT_REPORT_FOOTER: str = (
    "Colours are for visual orientation purposes only. "
    "Created with Colour{0}"
)
"""Report footer content."""

_VALUE_NOT_APPLICABLE: str = "N/A"


def _plot_report_header(axes: plt.Axes) -> plt.Axes:
    """
    Plot the report header, i.e. the title, on given axes.

    Parameters
    ----------
    axes
        Axes to add the report header to.

    Returns
    -------
    :class:`matplotlib.axes._axes.Axes`
        Axes the report header was added to.
    """

    axes.set_axis_off()
    axes.text(
        0.5,
        0.5,
        CONTENT_REPORT_HEADER,
        ha="center",
        va="center",
        size="x-large",
        weight="bold",
        zorder=CONSTANTS_COLOUR_STYLE.zorder.foreground_label,
    )

    return axes


def _plot_report_footer(axes: plt.Axes) -> plt.Axes:
    """
    Plot the report footer on given axes.

    Parameters
    ----------
    axes
        Axes to add the report footer to.

    Returns
    -------
    :class:`matplotlib.axes._axes.Axes`
        Axes the report footer was added to.
    """

    try:
        describe = describe_environment(print_callable=lambda x: x)[
            "colour-science.org"
        ]["colour"]
        version = f" {describe}."
    except Exception:  # pragma: no cover
        version = "."

    axes.set_axis_off()
    axes.text(
        0.5,
        0.5,
        CONTENT_REPORT_FOOTER.format(version),
        ha="center",
        va="center",
        size="small",
        zorder=CONSTANTS_COLOUR_STYLE.zorder.foreground_label,
    )


@override_style(**CONSTANTS_REPORT_STYLE)
def plot_single_sd_colour_rendition_report_full(
    sd: SpectralDistribution,
    source: Optional[str] = None,
    date: Optional[str] = None,
    manufacturer: Optional[str] = None,
    model: Optional[str] = None,
    notes: Optional[str] = None,
    report_size: Tuple = CONSTANT_REPORT_SIZE_FULL,
    report_row_height_ratios: Tuple = CONSTANT_REPORT_ROW_HEIGHT_RATIOS_FULL,
    report_box_padding: Optional[Dict] = None,
    **kwargs: Any,
) -> Tuple[plt.Figure, plt.Axes]:  # noqa: D405,D407,D410,D411
    """
    Generate the full *ANSI/IES TM-30-18 Colour Rendition Report* for given
    spectral distribution.

    Parameters
    ----------
    sd
        Spectral distribution of the emission source to generate the report
        for.
    source
        Emission source name, defaults to
        `colour.SpectralDistribution_IESTM2714.header.description` or
        `colour.SpectralDistribution_IESTM2714.name` properties value.
    date
        Emission source measurement date, defaults to
        `colour.SpectralDistribution_IESTM2714.header.report_date` property
        value.
    manufacturer
        Emission source manufacturer, defaults to
        `colour.SpectralDistribution_IESTM2714.header.manufacturer` property
        value.
    model
        Emission source model, defaults to
        `colour.SpectralDistribution_IESTM2714.header.catalog_number` property
        value.
    notes
        Notes pertaining to the emission source, defaults to
        `colour.SpectralDistribution_IESTM2714.header.comments` property
        value.
    report_size
        Report size, default to A4 paper size in inches.
    report_row_height_ratios
        Report size row height ratios.
    report_box_padding
        Report box padding, tries to define the padding around the figure and
        in-between the axes.

    Other Parameters
    ----------------
    kwargs
        {:func:`colour.plotting.artist`, :func:`colour.plotting.render`},
        See the documentation of the previously listed definitions.

    Returns
    -------
    :class:`tuple`
        Current figure and axes.

    Examples
    --------
    >>> from colour import SDS_ILLUMINANTS
    >>> sd = SDS_ILLUMINANTS['FL2']
    >>> plot_single_sd_colour_rendition_report_full(sd)
    ... # doctest: +ELLIPSIS
    (<Figure size ... with ... Axes>, <...AxesSubplot...>)

    .. image:: ../_static/Plotting_\
Plot_Single_SD_Colour_Rendition_Report_Full.png
        :align: center
        :alt: plot_single_sd_colour_rendition_report_full
    """

    report_box_padding = optional(
        report_box_padding, CONSTANT_REPORT_PADDING_FULL
    )

    specification: ColourQuality_Specification_ANSIIESTM3018 = cast(
        ColourQuality_Specification_ANSIIESTM3018,
        colour_fidelity_index_ANSIIESTM3018(sd, True),
    )

    sd = (
        SpectralDistribution_IESTM2714(data=sd, name=sd.name)
        if not isinstance(sd, SpectralDistribution_IESTM2714)
        else sd
    )

    NA = _VALUE_NOT_APPLICABLE

    source = optional(optional(source, sd.header.description), sd.name)
    date = optional(optional(date, sd.header.report_date), NA)
    manufacturer = optional(optional(manufacturer, sd.header.manufacturer), NA)
    model = optional(optional(model, sd.header.catalog_number), NA)
    notes = optional(optional(notes, sd.header.comments), NA)

    figure = plt.figure(figsize=report_size, constrained_layout=True)

    settings: Dict[str, Any] = dict(kwargs)
    settings["standalone"] = False
    settings["tight_layout"] = False

    gridspec_report = figure.add_gridspec(
        5, 1, height_ratios=report_row_height_ratios
    )

    # Title Row
    gridspec_title = gridspec_report[0].subgridspec(1, 1)
    axes_title = figure.add_subplot(gridspec_title[0])
    _plot_report_header(axes_title)

    # Description Rows & Columns
    gridspec_description = gridspec_report[1].subgridspec(1, 2)
    # Source & Date Column
    axes_source_date = figure.add_subplot(gridspec_description[0])
    axes_source_date.set_axis_off()
    axes_source_date.text(
        0.25,
        2 / 3,
        "Source: ",
        ha="right",
        va="center",
        size="medium",
        weight="bold",
    )
    axes_source_date.text(0.25, 2 / 3, source, va="center", size="medium")

    axes_source_date.text(
        0.25,
        1 / 3,
        "Date: ",
        ha="right",
        va="center",
        size="medium",
        weight="bold",
    )
    axes_source_date.text(0.25, 1 / 3, date, va="center", size="medium")

    # Manufacturer & Model Column
    axes_manufacturer_model = figure.add_subplot(gridspec_description[1])
    axes_manufacturer_model.set_axis_off()
    axes_manufacturer_model.text(
        0.25,
        2 / 3,
        "Manufacturer: ",
        ha="right",
        va="center",
        size="medium",
        weight="bold",
    )
    axes_manufacturer_model.text(
        0.25, 2 / 3, manufacturer, va="center", size="medium"
    )

    axes_manufacturer_model.text(
        0.25,
        1 / 3,
        "Model: ",
        ha="right",
        va="center",
        size="medium",
        weight="bold",
    )
    axes_manufacturer_model.text(
        0.25, 1 / 3, model, va="center", size="medium"
    )

    # Main Figures Rows & Columns
    gridspec_figures = gridspec_report[2].subgridspec(
        4, 2, height_ratios=[1, 1, 1, 1.5]
    )
    axes_spectra = figure.add_subplot(gridspec_figures[0, 0])
    plot_spectra_ANSIIESTM3018(specification, axes=axes_spectra, **settings)

    axes_vector_graphics = figure.add_subplot(gridspec_figures[1:3, 0])
    plot_colour_vector_graphic(
        specification, axes=axes_vector_graphics, **settings
    )

    axes_chroma_shifts = figure.add_subplot(gridspec_figures[0, 1])
    plot_local_chroma_shifts(
        specification, axes=axes_chroma_shifts, **settings
    )

    axes_hue_shifts = figure.add_subplot(gridspec_figures[1, 1])
    plot_local_hue_shifts(specification, axes=axes_hue_shifts, **settings)

    axes_colour_fidelities = figure.add_subplot(gridspec_figures[2, 1])
    plot_local_colour_fidelities(
        specification, axes=axes_colour_fidelities, x_ticker=True, **settings
    )

    # Colour Fidelity Indexes Row
    axes_colour_fidelity_indexes = figure.add_subplot(gridspec_figures[3, :])
    plot_colour_fidelity_indexes(
        specification, axes=axes_colour_fidelity_indexes, **settings
    )

    # Notes & Chromaticities / CRI Row and Columns
    gridspec_notes_chromaticities_CRI = gridspec_report[3].subgridspec(1, 2)
    axes_notes = figure.add_subplot(gridspec_notes_chromaticities_CRI[0])
    axes_notes.set_axis_off()
    axes_notes.text(
        0.25,
        1,
        "Notes: ",
        ha="right",
        va="center",
        size="medium",
        weight="bold",
    )
    axes_notes.text(0.25, 1, notes, va="center", size="medium")
    gridspec_chromaticities_CRI = gridspec_notes_chromaticities_CRI[
        1
    ].subgridspec(1, 2)

    XYZ = sd_to_XYZ(specification.sd_test)
    xy = XYZ_to_xy(XYZ)
    Luv = XYZ_to_Luv(XYZ, xy)
    uv_p = Luv_to_uv(Luv, xy)

    gridspec_chromaticities = gridspec_chromaticities_CRI[0].subgridspec(1, 1)
    axes_chromaticities = figure.add_subplot(gridspec_chromaticities[0])
    axes_chromaticities.set_axis_off()
    axes_chromaticities.text(
        0.5,
        4 / 5,
        f"$x$ {xy[0]:.4f}",
        ha="center",
        va="center",
        size="medium",
        weight="bold",
    )

    axes_chromaticities.text(
        0.5,
        3 / 5,
        f"$y$ {xy[1]:.4f}",
        ha="center",
        va="center",
        size="medium",
        weight="bold",
    )

    axes_chromaticities.text(
        0.5,
        2 / 5,
        f"$u'$ {uv_p[0]:.4f}",
        ha="center",
        va="center",
        size="medium",
        weight="bold",
    )

    axes_chromaticities.text(
        0.5,
        1 / 5,
        f"$v'$ {uv_p[1]:.4f}",
        ha="center",
        va="center",
        size="medium",
        weight="bold",
    )

    gridspec_CRI = gridspec_chromaticities_CRI[1].subgridspec(1, 1)

    CRI_spec: ColourRendering_Specification_CRI = cast(
        ColourRendering_Specification_CRI,
        colour_rendering_index(specification.sd_test, additional_data=True),
    )

    axes_CRI = figure.add_subplot(gridspec_CRI[0])
    axes_CRI.set_xticks([])
    axes_CRI.set_yticks([])
    axes_CRI.text(
        0.5,
        4 / 5,
        "CIE 13.31-1995",
        ha="center",
        va="center",
        size="medium",
        weight="bold",
    )

    axes_CRI.text(
        0.5,
        3 / 5,
        "(CRI)",
        ha="center",
        va="center",
        size="medium",
        weight="bold",
    )

    axes_CRI.text(
        0.5,
        2 / 5,
        f"$R_a$ {float(CRI_spec.Q_a):.0f}",
        ha="center",
        va="center",
        size="medium",
        weight="bold",
    )

    axes_CRI.text(
        0.5,
        1 / 5,
        f"$R_9$ {float(CRI_spec.Q_as[8].Q_a):.0f}",
        ha="center",
        va="center",
        size="medium",
        weight="bold",
    )

    gridspec_footer = gridspec_report[4].subgridspec(1, 1)
    axes_footer = figure.add_subplot(gridspec_footer[0])
    _plot_report_footer(axes_footer)

    figure.set_constrained_layout_pads(**report_box_padding)

    settings = dict(kwargs)
    settings["tight_layout"] = False

    return render(**settings)


@override_style(**CONSTANTS_REPORT_STYLE)
def plot_single_sd_colour_rendition_report_intermediate(
    sd: SpectralDistribution,
    report_size: Tuple = CONSTANT_REPORT_SIZE_INTERMEDIATE,
    report_row_height_ratios: Tuple = (
        CONSTANT_REPORT_ROW_HEIGHT_RATIOS_INTERMEDIATE
    ),
    report_box_padding: Optional[Dict] = None,
    **kwargs: Any,
) -> Tuple[plt.Figure, plt.Axes]:
    """
    Generate the intermediate *ANSI/IES TM-30-18 Colour Rendition Report* for
    given spectral distribution.

    Parameters
    ----------
    sd
        Spectral distribution of the emission source to generate the report
        for.
    report_size
        Report size, default to A4 paper size in inches.
    report_row_height_ratios
        Report size row height ratios.
    report_box_padding
        Report box padding, tries to define the padding around the figure and
        in-between the axes.

    Other Parameters
    ----------------
    kwargs
        {:func:`colour.plotting.artist`, :func:`colour.plotting.render`},
        See the documentation of the previously listed definitions.

    Returns
    -------
    :class:`tuple`
        Current figure and axes.

    Examples
    --------
    >>> from colour import SDS_ILLUMINANTS
    >>> sd = SDS_ILLUMINANTS['FL2']
    >>> plot_single_sd_colour_rendition_report_intermediate(sd)
    ... # doctest: +ELLIPSIS
    (<Figure size ... with ... Axes>, <...AxesSubplot...>)

    .. image:: ../_static/Plotting_\
Plot_Single_SD_Colour_Rendition_Report_Intermediate.png
        :align: center
        :alt: plot_single_sd_colour_rendition_report_intermediate
    """

    report_box_padding = optional(
        report_box_padding, CONSTANT_REPORT_PADDING_INTERMEDIATE
    )

    specification: ColourQuality_Specification_ANSIIESTM3018 = cast(
        ColourQuality_Specification_ANSIIESTM3018,
        colour_fidelity_index_ANSIIESTM3018(sd, True),
    )

    figure = plt.figure(figsize=report_size, constrained_layout=True)

    settings: Dict[str, Any] = dict(kwargs)
    settings["standalone"] = False
    settings["tight_layout"] = False

    gridspec_report = figure.add_gridspec(
        3, 1, height_ratios=report_row_height_ratios
    )

    # Title Row
    gridspec_title = gridspec_report[0].subgridspec(1, 1)
    axes_title = figure.add_subplot(gridspec_title[0])
    _plot_report_header(axes_title)

    # Main Figures Rows & Columns
    gridspec_figures = gridspec_report[1].subgridspec(2, 2)

    axes_vector_graphics = figure.add_subplot(gridspec_figures[0:2, 0])
    plot_colour_vector_graphic(
        specification, axes=axes_vector_graphics, **settings
    )

    axes_chroma_shifts = figure.add_subplot(gridspec_figures[0, 1])
    plot_local_chroma_shifts(
        specification, axes=axes_chroma_shifts, **settings
    )

    axes_hue_shifts = figure.add_subplot(gridspec_figures[1, 1])
    plot_local_hue_shifts(
        specification, axes=axes_hue_shifts, x_ticker=True, **settings
    )

    gridspec_footer = gridspec_report[2].subgridspec(1, 1)
    axes_footer = figure.add_subplot(gridspec_footer[0])
    _plot_report_footer(axes_footer)

    figure.set_constrained_layout_pads(**report_box_padding)

    settings = dict(kwargs)
    settings["tight_layout"] = False

    return render(**settings)


def plot_single_sd_colour_rendition_report_simple(
    sd: SpectralDistribution,
    report_size: Tuple = CONSTANT_REPORT_SIZE_SIMPLE,
    report_row_height_ratios: Tuple = CONSTANT_REPORT_ROW_HEIGHT_RATIOS_SIMPLE,
    report_box_padding: Optional[Dict] = None,
    **kwargs: Any,
) -> Tuple[plt.Figure, plt.Axes]:
    """
    Generate the simple *ANSI/IES TM-30-18 Colour Rendition Report* for given
    spectral distribution.

    Parameters
    ----------
    sd
        Spectral distribution of the emission source to generate the report
        for.
    report_size
        Report size, default to A4 paper size in inches.
    report_row_height_ratios
        Report size row height ratios.
    report_box_padding
        Report box padding, tries to define the padding around the figure and
        in-between the axes.

    Other Parameters
    ----------------
    kwargs
        {:func:`colour.plotting.artist`, :func:`colour.plotting.render`},
        See the documentation of the previously listed definitions.

    Returns
    -------
    :class:`tuple`
        Current figure and axes.

    Examples
    --------
    >>> from colour import SDS_ILLUMINANTS
    >>> sd = SDS_ILLUMINANTS['FL2']
    >>> plot_single_sd_colour_rendition_report_simple(sd)
    ... # doctest: +ELLIPSIS
    (<Figure size ... with ... Axes>, <...AxesSubplot...>)

    .. image:: ../_static/Plotting_\
Plot_Single_SD_Colour_Rendition_Report_Simple.png
        :align: center
        :alt: plot_single_sd_colour_rendition_report_simple
    """

    report_box_padding = optional(
        report_box_padding, CONSTANT_REPORT_PADDING_SIMPLE
    )

    specification: ColourQuality_Specification_ANSIIESTM3018 = cast(
        ColourQuality_Specification_ANSIIESTM3018,
        colour_fidelity_index_ANSIIESTM3018(sd, True),
    )

    figure = plt.figure(figsize=report_size, constrained_layout=True)

    settings: Dict[str, Any] = dict(kwargs)
    settings["standalone"] = False
    settings["tight_layout"] = False

    gridspec_report = figure.add_gridspec(
        3, 1, height_ratios=report_row_height_ratios
    )

    # Title Row
    gridspec_title = gridspec_report[0].subgridspec(1, 1)
    axes_title = figure.add_subplot(gridspec_title[0])
    _plot_report_header(axes_title)

    # Main Figures Rows & Columns
    gridspec_figures = gridspec_report[1].subgridspec(1, 1)

    axes_vector_graphics = figure.add_subplot(gridspec_figures[0, 0])
    plot_colour_vector_graphic(
        specification, axes=axes_vector_graphics, **settings
    )

    gridspec_footer = gridspec_report[2].subgridspec(1, 1)
    axes_footer = figure.add_subplot(gridspec_footer[0])
    _plot_report_footer(axes_footer)

    figure.set_constrained_layout_pads(**report_box_padding)

    settings = dict(kwargs)
    settings["tight_layout"] = False

    return render(**settings)


def plot_single_sd_colour_rendition_report(
    sd: SpectralDistribution,
    method: Union[Literal["Full", "Intermediate", "Simple"], str] = "Full",
    **kwargs: Any,
) -> Tuple[plt.Figure, plt.Axes]:
    """
    Generate the *ANSI/IES TM-30-18 Colour Rendition Report* for given
    spectral distribution according to given method.

    Parameters
    ----------
    sd
        Spectral distribution of the emission source to generate the report
        for.
    method
        Report plotting method.

    Other Parameters
    ----------------
    kwargs
        {:func:`colour.plotting.artist`, :func:`colour.plotting.render`,
        :func:`colour.plotting.tm3018.\
plot_single_sd_colour_rendition_report_full`, :func:`colour.plotting.tm3018.\
plot_single_sd_colour_rendition_report_intermediate`, \
:func:`colour.plotting.tm3018.plot_single_sd_colour_rendition_report_simple`}
        See the documentation of the previously listed definitions.

    Returns
    -------
    :class:`tuple`
        Current figure and axes.

    Examples
    --------
    >>> from colour import SDS_ILLUMINANTS
    >>> sd = SDS_ILLUMINANTS['FL2']
    >>> plot_single_sd_colour_rendition_report(sd)
    ... # doctest: +ELLIPSIS
    (<Figure size ... with ... Axes>, <...AxesSubplot...>)

    .. image:: ../_static/Plotting_\
Plot_Single_SD_Colour_Rendition_Report_Full.png
        :align: center
        :alt: plot_single_sd_colour_rendition_report_full

    >>> plot_single_sd_colour_rendition_report(sd, 'Intermediate')
    ... # doctest: +ELLIPSIS
    (<Figure size ... with ... Axes>, <...AxesSubplot...>)

    .. image:: ../_static/Plotting_\
Plot_Single_SD_Colour_Rendition_Report_Intermediate.png
        :align: center
        :alt: plot_single_sd_colour_rendition_report_intermediate

    >>> plot_single_sd_colour_rendition_report(sd, 'Simple')
    ... # doctest: +ELLIPSIS
    (<Figure size ... with ... Axes>, <...AxesSubplot...>)

    .. image:: ../_static/Plotting_\
Plot_Single_SD_Colour_Rendition_Report_Simple.png
        :align: center
        :alt: plot_single_sd_colour_rendition_report_simple
    """

    method = validate_method(method, ["Full", "Intermediate", "Simple"])

    if method == "full":
        return plot_single_sd_colour_rendition_report_full(sd, **kwargs)
    elif method == "intermediate":
        return plot_single_sd_colour_rendition_report_intermediate(
            sd, **kwargs
        )
    else:  # method == 'simple'
        return plot_single_sd_colour_rendition_report_simple(sd, **kwargs)



"""
Prismatic Colourspace
=====================

Defines the *Prismatic* colourspace transformations:

-   :func:`colour.RGB_to_Prismatic`
-   :func:`colour.Prismatic_to_RGB`

References
----------
-   :cite:`Shirley2015a` : Shirley, P., & Hart, D. (2015). The prismatic color
    space for rgb computations (pp. 2-7).
"""

from __future__ import annotations

import numpy as np

from colour.hints import ArrayLike, NDArray
from colour.utilities import from_range_1, to_domain_1, tsplit, tstack

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "RGB_to_Prismatic",
    "Prismatic_to_RGB",
]


def RGB_to_Prismatic(RGB: ArrayLike) -> NDArray:
    """
    Convert from *RGB* colourspace to *Prismatic* :math:`L\\rho\\gamma\\beta`
    colourspace array.

    Parameters
    ----------
    RGB
        *RGB* colourspace array.

    Returns
    -------
    :class:`numpy.ndarray`
        *Prismatic* :math:`L\\rho\\gamma\\beta` colourspace array.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``RGB``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``Lrgb``   | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Shirley2015a`

    Examples
    --------
    >>> RGB = np.array([0.25, 0.50, 0.75])
    >>> RGB_to_Prismatic(RGB)  # doctest: +ELLIPSIS
    array([ 0.75...   ,  0.1666666...,  0.3333333...,  0.5...   ])

    Adjusting saturation of given *RGB* colourspace array:
    >>> saturation = 0.5
    >>> Lrgb = RGB_to_Prismatic(RGB)
    >>> Lrgb[..., 1:] = 1 / 3 + saturation * (Lrgb[..., 1:] - 1 / 3)
    >>> Prismatic_to_RGB(Lrgb)  # doctest: +ELLIPSIS
    array([ 0.45...,  0.6...,  0.75...])
    """

    RGB = to_domain_1(RGB)

    L = np.max(RGB, axis=-1)
    s = np.sum(RGB, axis=-1)[..., np.newaxis]
    one_s = 1 / s
    # Handling zero-division *NaNs*.
    one_s[s == 0] = 0
    r, g, b = tsplit(one_s * RGB)

    Lrgb = tstack([L, r, g, b])

    return from_range_1(Lrgb)


def Prismatic_to_RGB(Lrgb: ArrayLike) -> NDArray:
    """
    Convert from *Prismatic* :math:`L\\rho\\gamma\\beta` colourspace array to
    *RGB* colourspace.

    Parameters
    ----------
    Lrgb
        *Prismatic* :math:`L\\rho\\gamma\\beta` colourspace array.

    Returns
    -------
    :class:`numpy.ndarray`
        *RGB* colourspace array.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``Lrgb``   | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``RGB``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Shirley2015a`

    Examples
    --------
    >>> Lrgb = np.array([0.75000000, 0.16666667, 0.33333333, 0.50000000])
    >>> Prismatic_to_RGB(Lrgb)  # doctest: +ELLIPSIS
    array([ 0.25...   ,  0.4999999...,  0.75...  ])
    """

    Lrgb = to_domain_1(Lrgb)

    rgb = Lrgb[..., 1:]
    m = np.max(rgb, axis=-1)[..., np.newaxis]
    RGB = Lrgb[..., 0][..., np.newaxis] / m
    # Handling zero-division *NaNs*.
    RGB[m == 0] = 0
    RGB = RGB * rgb

    return from_range_1(RGB)



"""
Documentation
=============

Defines the documentation related objects.
"""

from __future__ import annotations

from colour.hints import Boolean

import os

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "DocstringDict",
    "DocstringFloat",
    "DocstringInt",
    "DocstringText",
    "DocstringTuple",
    "is_documentation_building",
]


class DocstringDict(dict):
    """
    A :class:`dict` sub-class that allows settings a docstring to :class:`dict`
    instances.
    """

    pass


class DocstringFloat(float):
    """
    A :class:`float` sub-class that allows settings a docstring to
    :class:`float` instances.
    """

    pass


class DocstringInt(int):
    """
    A :class:`numpy.integer` sub-class that allows settings a docstring to
    :class:`numpy.integer` instances.
    """

    pass


class DocstringText(str):
    """
    A :class:`str` sub-class that allows settings a docstring to
    :class:`str` instances.
    """

    pass


class DocstringTuple(tuple):
    """
    A :class:`tuple` sub-class that allows settings a docstring to
    :class:`tuple` instances.
    """

    pass


def is_documentation_building() -> Boolean:
    """
    Return whether the documentation is being built by checking whether the
    *READTHEDOCS* or *COLOUR_SCIENCE__DOCUMENTATION_BUILD* environment
    variables are defined, their value is not accounted for.

    Returns
    -------
    :class:`bool`
        Whether the documentation is being built.

    Examples
    --------
    >>> is_documentation_building()
    False
    >>> os.environ['READTHEDOCS'] = 'True'
    >>> is_documentation_building()
    True
    >>> os.environ['READTHEDOCS'] = 'False'
    >>> is_documentation_building()
    True
    >>> del os.environ['READTHEDOCS']
    >>> is_documentation_building()
    False
    >>> os.environ['COLOUR_SCIENCE__DOCUMENTATION_BUILD'] = 'Yes'
    >>> is_documentation_building()
    True
    >>> del os.environ['COLOUR_SCIENCE__DOCUMENTATION_BUILD']
    """

    return bool(
        os.environ.get("READTHEDOCS")
        or os.environ.get("COLOUR_SCIENCE__DOCUMENTATION_BUILD")
    )



"""
Defines the unit tests for the
:mod:`colour.algebra.coordinates.transformations` module.
"""

import numpy as np
import unittest
from itertools import permutations

from colour.algebra import (
    cartesian_to_spherical,
    spherical_to_cartesian,
    cartesian_to_polar,
    polar_to_cartesian,
    cartesian_to_cylindrical,
    cylindrical_to_cartesian,
)
from colour.utilities import ignore_numpy_errors

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestCartesianToSpherical",
    "TestSphericalToCartesian",
    "TestCartesianToPolar",
    "TestPolarToCartesian",
    "TestCartesianToCylindrical",
    "TestCylindricalToCartesian",
]


class TestCartesianToSpherical(unittest.TestCase):
    """
    Define :func:`colour.algebra.coordinates.transformations.\
cartesian_to_spherical` definition unit tests methods.
    """

    def test_cartesian_to_spherical(self):
        """
        Test :func:`colour.algebra.coordinates.transformations.\
cartesian_to_spherical` definition.
        """

        np.testing.assert_almost_equal(
            cartesian_to_spherical(np.array([3, 1, 6])),
            np.array([6.78232998, 0.48504979, 0.32175055]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            cartesian_to_spherical(np.array([-1, 9, 16])),
            np.array([18.38477631, 0.51501513, 1.68145355]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            cartesian_to_spherical(np.array([6.3434, -0.9345, 18.5675])),
            np.array([19.64342307, 0.33250603, -0.14626640]),
            decimal=7,
        )

    def test_n_dimensional_cartesian_to_spherical(self):
        """
        Test :func:`colour.algebra.coordinates.transformations.\
cartesian_to_spherical` definition n-dimensional arrays support.
        """

        a_i = np.array([3, 1, 6])
        a_o = cartesian_to_spherical(a_i)

        a_i = np.tile(a_i, (6, 1))
        a_o = np.tile(a_o, (6, 1))
        np.testing.assert_almost_equal(
            cartesian_to_spherical(a_i), a_o, decimal=7
        )

        a_i = np.reshape(a_i, (2, 3, 3))
        a_o = np.reshape(a_o, (2, 3, 3))
        np.testing.assert_almost_equal(
            cartesian_to_spherical(a_i), a_o, decimal=7
        )

    @ignore_numpy_errors
    def test_nan_cartesian_to_spherical(self):
        """
        Test :func:`colour.algebra.coordinates.transformations.\
cartesian_to_spherical` definition nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            a_i = np.array(case)
            cartesian_to_spherical(a_i)


class TestSphericalToCartesian(unittest.TestCase):
    """
    Define :func:`colour.algebra.coordinates.transformations.\
spherical_to_cartesian` definition unit tests methods.
    """

    def test_spherical_to_cartesian(self):
        """
        Test :func:`colour.algebra.coordinates.transformations.\
spherical_to_cartesian` definition.
        """

        np.testing.assert_almost_equal(
            spherical_to_cartesian(
                np.array([6.78232998, 0.48504979, 0.32175055])
            ),
            np.array([3.00000000, 0.99999999, 6.00000000]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            spherical_to_cartesian(
                np.array([18.38477631, 0.51501513, 1.68145355])
            ),
            np.array([-1.00000003, 9.00000007, 15.99999996]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            spherical_to_cartesian(
                np.array([19.64342307, 0.33250603, -0.14626640])
            ),
            np.array([6.34339996, -0.93449999, 18.56750001]),
            decimal=7,
        )

    def test_n_dimensional_spherical_to_cartesian(self):
        """
        Test :func:`colour.algebra.coordinates.transformations.\
spherical_to_cartesian` definition n-dimensional arrays support.
        """

        a_i = np.array([6.78232998, 0.48504979, 0.32175055])
        a_o = spherical_to_cartesian(a_i)

        a_i = np.tile(a_i, (6, 1))
        a_o = np.tile(a_o, (6, 1))
        np.testing.assert_almost_equal(
            spherical_to_cartesian(a_i), a_o, decimal=7
        )

        a_i = np.reshape(a_i, (2, 3, 3))
        a_o = np.reshape(a_o, (2, 3, 3))
        np.testing.assert_almost_equal(
            spherical_to_cartesian(a_i), a_o, decimal=7
        )

    @ignore_numpy_errors
    def test_nan_spherical_to_cartesian(self):
        """
        Test :func:`colour.algebra.coordinates.transformations.\
spherical_to_cartesian` definition nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            a_i = np.array(case)
            spherical_to_cartesian(a_i)


class TestCartesianToPolar(unittest.TestCase):
    """
    Define :func:`colour.algebra.coordinates.transformations.\
cartesian_to_polar` definition unit tests methods.
    """

    def test_cartesian_to_polar(self):
        """
        Test :func:`colour.algebra.coordinates.transformations.\
cartesian_to_polar` definition.
        """

        np.testing.assert_almost_equal(
            cartesian_to_polar(np.array([3, 1])),
            np.array([3.16227766, 0.32175055]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            cartesian_to_polar(np.array([-1, 9])),
            np.array([9.05538514, 1.68145355]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            cartesian_to_polar(np.array([6.3434, -0.9345])),
            np.array([6.41186508, -0.14626640]),
            decimal=7,
        )

    def test_n_dimensional_cartesian_to_polar(self):
        """
        Test :func:`colour.algebra.coordinates.transformations.\
cartesian_to_polar` definition n-dimensional arrays support.
        """

        a_i = np.array([3, 1])
        a_o = cartesian_to_polar(a_i)

        a_i = np.tile(a_i, (6, 1))
        a_o = np.tile(a_o, (6, 1))
        np.testing.assert_almost_equal(cartesian_to_polar(a_i), a_o, decimal=7)

        a_i = np.reshape(a_i, (2, 3, 2))
        a_o = np.reshape(a_o, (2, 3, 2))
        np.testing.assert_almost_equal(cartesian_to_polar(a_i), a_o, decimal=7)

    @ignore_numpy_errors
    def test_nan_cartesian_to_polar(self):
        """
        Test :func:`colour.algebra.coordinates.transformations.\
cartesian_to_polar` definition nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=2))
        for case in cases:
            a_i = np.array(case)
            cartesian_to_polar(a_i)


class TestPolarToCartesian(unittest.TestCase):
    """
    Define :func:`colour.algebra.coordinates.transformations.\
polar_to_cartesian` definition unit tests methods.
    """

    def test_polar_to_cartesian(self):
        """
        Test :func:`colour.algebra.coordinates.transformations.\
polar_to_cartesian` definition.
        """

        np.testing.assert_almost_equal(
            polar_to_cartesian(np.array([0.32175055, 1.08574654])),
            np.array([0.15001697, 0.28463718]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            polar_to_cartesian(np.array([1.68145355, 1.05578119])),
            np.array([0.82819662, 1.46334425]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            polar_to_cartesian(np.array([-0.14626640, 1.23829030])),
            np.array([-0.04774323, -0.13825500]),
            decimal=7,
        )

    def test_n_dimensional_polar_to_cartesian(self):
        """
        Test :func:`colour.algebra.coordinates.transformations.\
polar_to_cartesian` definition n-dimensional arrays support.
        """

        a_i = np.array([3.16227766, 0.32175055])
        a_o = polar_to_cartesian(a_i)

        a_i = np.tile(a_i, (6, 1))
        a_o = np.tile(a_o, (6, 1))
        np.testing.assert_almost_equal(polar_to_cartesian(a_i), a_o, decimal=7)

        a_i = np.reshape(a_i, (2, 3, 2))
        a_o = np.reshape(a_o, (2, 3, 2))
        np.testing.assert_almost_equal(polar_to_cartesian(a_i), a_o, decimal=7)

    @ignore_numpy_errors
    def test_nan_polar_to_cartesian(self):
        """
        Test :func:`colour.algebra.coordinates.transformations.\
polar_to_cartesian` definition nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=2))
        for case in cases:
            a_i = np.array(case)
            polar_to_cartesian(a_i)


class TestCartesianToCylindrical(unittest.TestCase):
    """
    Define :func:`colour.algebra.coordinates.transformations.\
cartesian_to_cylindrical` definition unit tests methods.
    """

    def test_cartesian_to_cylindrical(self):
        """
        Test :func:`colour.algebra.coordinates.transformations.\
cartesian_to_cylindrical` definition.
        """

        np.testing.assert_almost_equal(
            cartesian_to_cylindrical(np.array([3, 1, 6])),
            np.array([3.16227766, 0.32175055, 6.00000000]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            cartesian_to_cylindrical(np.array([-1, 9, 16])),
            np.array([9.05538514, 1.68145355, 16.00000000]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            cartesian_to_cylindrical(np.array([6.3434, -0.9345, 18.5675])),
            np.array([6.41186508, -0.14626640, 18.56750000]),
            decimal=7,
        )

    def test_n_dimensional_cartesian_to_cylindrical(self):
        """
        Test :func:`colour.algebra.coordinates.transformations.\
cartesian_to_cylindrical` definition n-dimensional arrays support.
        """

        a_i = np.array([3, 1, 6])
        a_o = cartesian_to_cylindrical(a_i)

        a_i = np.tile(a_i, (6, 1))
        a_o = np.tile(a_o, (6, 1))
        np.testing.assert_almost_equal(
            cartesian_to_cylindrical(a_i), a_o, decimal=7
        )

        a_i = np.reshape(a_i, (2, 3, 3))
        a_o = np.reshape(a_o, (2, 3, 3))
        np.testing.assert_almost_equal(
            cartesian_to_cylindrical(a_i), a_o, decimal=7
        )

    @ignore_numpy_errors
    def test_nan_cartesian_to_cylindrical(self):
        """
        Test :func:`colour.algebra.coordinates.transformations.\
cartesian_to_cylindrical` definition nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            a_i = np.array(case)
            cartesian_to_cylindrical(a_i)


class TestCylindricalToCartesian(unittest.TestCase):
    """
    Define :func:`colour.algebra.coordinates.transformations.\
cylindrical_to_cartesian` definition unit tests methods.
    """

    def test_cylindrical_to_cartesian(self):
        """
        Test :func:`colour.algebra.coordinates.transformations.\
cylindrical_to_cartesian` definition.
        """

        np.testing.assert_almost_equal(
            cylindrical_to_cartesian(
                np.array([0.32175055, 1.08574654, 6.78232998])
            ),
            np.array([0.15001697, 0.28463718, 6.78232998]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            cylindrical_to_cartesian(
                np.array([1.68145355, 1.05578119, 18.38477631])
            ),
            np.array([0.82819662, 1.46334425, 18.38477631]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            cylindrical_to_cartesian(
                np.array([-0.14626640, 1.23829030, 19.64342307])
            ),
            np.array([-0.04774323, -0.13825500, 19.64342307]),
            decimal=7,
        )

    def test_n_dimensional_cylindrical_to_cartesian(self):
        """
        Test :func:`colour.algebra.coordinates.transformations.\
cylindrical_to_cartesian` definition n-dimensional arrays support.
        """

        a_i = np.array([3.16227766, 0.32175055, 6.00000000])
        a_o = cylindrical_to_cartesian(a_i)

        a_i = np.tile(a_i, (6, 1))
        a_o = np.tile(a_o, (6, 1))
        np.testing.assert_almost_equal(
            cylindrical_to_cartesian(a_i), a_o, decimal=7
        )

        a_i = np.reshape(a_i, (2, 3, 3))
        a_o = np.reshape(a_o, (2, 3, 3))
        np.testing.assert_almost_equal(
            cylindrical_to_cartesian(a_i), a_o, decimal=7
        )

    @ignore_numpy_errors
    def test_nan_cylindrical_to_cartesian(self):
        """
        Test :func:`colour.algebra.coordinates.transformations.\
cylindrical_to_cartesian` definition nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            a_i = np.array(case)
            cylindrical_to_cartesian(a_i)


if __name__ == "__main__":
    unittest.main()



"""Defines the unit tests for the :mod:`colour.recovery.jakob2019` module."""

import numpy as np
import os
import shutil
import tempfile
import unittest

from colour.characterisation import SDS_COLOURCHECKERS
from colour.colorimetry import handle_spectral_arguments, sd_to_XYZ
from colour.difference import JND_CIE1976, delta_E_CIE1976
from colour.models import (
    RGB_COLOURSPACE_sRGB,
    RGB_to_XYZ,
    XYZ_to_Lab,
    XYZ_to_xy,
)
from colour.recovery.jakob2019 import (
    XYZ_to_sd_Jakob2019,
    sd_Jakob2019,
    error_function,
    dimensionalise_coefficients,
    SPECTRAL_SHAPE_JAKOB2019,
    LUT3D_Jakob2019,
)
from colour.utilities import domain_range_scale, full, ones, zeros

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestErrorFunction",
    "TestXYZ_to_sd_Jakob2019",
    "TestLUT3D_Jakob2019",
]


class TestErrorFunction(unittest.TestCase):
    """
    Define :func:`colour.recovery.jakob2019.error_function` definition unit
    tests methods.
    """

    def setUp(self):
        """Initialise the common tests attributes."""

        self._shape = SPECTRAL_SHAPE_JAKOB2019
        self._cmfs, self._sd_D65 = handle_spectral_arguments(
            shape_default=self._shape
        )
        self._XYZ_D65 = sd_to_XYZ(self._sd_D65)
        self._xy_D65 = XYZ_to_xy(self._XYZ_D65)

        self._Lab_e = np.array([72, -20, 61])

    def test_intermediates(self):
        """
        Test intermediate results of
        :func:`colour.recovery.jakob2019.error_function` with
        :func:`colour.sd_to_XYZ`, :func:`colour.XYZ_to_Lab` and checks if the
        error is computed correctly by comparing it with
        :func:`colour.difference.delta_E_CIE1976`.
        """

        # Quoted names refer to colours from ColorChecker N Ohta (using D65).
        coefficient_list = [
            np.array([0, 0, 0]),  # 50% gray
            np.array([0, 0, -1e9]),  # Pure black
            np.array([0, 0, +1e9]),  # Pure white
            np.array([1e9, -1e9, 2.1e8]),  # A pathological example
            np.array([2.2667394, -7.6313081, 1.03185]),  # 'blue'
            np.array([-31.377077, 26.810094, -6.1139927]),  # 'green'
            np.array([25.064246, -16.072039, 0.10431365]),  # 'red'
            np.array([-19.325667, 22.242319, -5.8144924]),  # 'yellow'
            np.array([21.909902, -17.227963, 2.142351]),  # 'magenta'
            np.array([-15.864009, 8.6735071, -1.4012552]),  # 'cyan'
        ]

        for coefficients in coefficient_list:
            error, _derror, R, XYZ, Lab = error_function(
                coefficients,
                self._Lab_e,
                self._cmfs,
                self._sd_D65,
                additional_data=True,
            )

            sd = sd_Jakob2019(
                dimensionalise_coefficients(coefficients, self._shape),
                self._shape,
            )

            sd_XYZ = sd_to_XYZ(sd, self._cmfs, self._sd_D65) / 100
            sd_Lab = XYZ_to_Lab(XYZ, self._xy_D65)
            error_reference = delta_E_CIE1976(self._Lab_e, Lab)

            np.testing.assert_allclose(sd.values, R, atol=1e-14)
            np.testing.assert_allclose(XYZ, sd_XYZ, atol=1e-14)

            self.assertLess(abs(error_reference - error), JND_CIE1976 / 100)
            self.assertLess(delta_E_CIE1976(Lab, sd_Lab), JND_CIE1976 / 100)

    def test_derivatives(self):
        """
        Test the gradients computed using closed-form expressions of the
        derivatives with finite difference approximations.
        """

        samples = np.linspace(-10, 10, 1000)
        ds = samples[1] - samples[0]

        # Vary one coefficient at a time, keeping the others fixed to 1.
        for coefficient_i in range(3):
            errors = np.empty_like(samples)
            derrors = np.empty_like(samples)

            for i, sample in enumerate(samples):
                coefficients = ones(3)
                coefficients[coefficient_i] = sample

                error, derror = error_function(
                    coefficients, self._Lab_e, self._cmfs, self._sd_D65
                )

                errors[i] = error
                derrors[i] = derror[coefficient_i]

            staggered_derrors = (derrors[:-1] + derrors[1:]) / 2
            approximate_derrors = np.diff(errors) / ds

            # The approximated derivatives aren't too accurate, so tolerances
            # have to be rather loose.
            np.testing.assert_allclose(
                staggered_derrors, approximate_derrors, atol=1e-3, rtol=1e-2
            )


class TestXYZ_to_sd_Jakob2019(unittest.TestCase):
    """
    Define :func:`colour.recovery.jakob2019.XYZ_to_sd_Jakob2019` definition
    unit tests methods.
    """

    def setUp(self):
        """Initialise the common tests attributes."""

        self._shape = SPECTRAL_SHAPE_JAKOB2019
        self._cmfs, self._sd_D65 = handle_spectral_arguments(
            shape_default=self._shape
        )

    def test_XYZ_to_sd_Jakob2019(self):
        """Test :func:`colour.recovery.jakob2019.XYZ_to_sd_Jakob2019` definition."""

        # Tests the round-trip with values of a colour checker.
        for name, sd in SDS_COLOURCHECKERS["ColorChecker N Ohta"].items():
            XYZ = sd_to_XYZ(sd, self._cmfs, self._sd_D65) / 100

            _recovered_sd, error = XYZ_to_sd_Jakob2019(
                XYZ, self._cmfs, self._sd_D65, additional_data=True
            )

            if error > JND_CIE1976 / 100:  # pragma: no cover
                self.fail(f"Delta E for '{name}' is {error}!")

    def test_domain_range_scale_XYZ_to_sd_Jakob2019(self):
        """
        Test :func:`colour.recovery.jakob2019.XYZ_to_sd_Jakob2019` definition
        domain and range scale support.
        """

        XYZ_i = np.array([0.20654008, 0.12197225, 0.05136952])
        XYZ_o = sd_to_XYZ(
            XYZ_to_sd_Jakob2019(XYZ_i, self._cmfs, self._sd_D65),
            self._cmfs,
            self._sd_D65,
        )

        d_r = (("reference", 1, 1), ("1", 1, 0.01), ("100", 100, 1))
        for scale, factor_a, factor_b in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    sd_to_XYZ(
                        XYZ_to_sd_Jakob2019(
                            XYZ_i * factor_a, self._cmfs, self._sd_D65
                        ),
                        self._cmfs,
                        self._sd_D65,
                    ),
                    XYZ_o * factor_b,
                    decimal=7,
                )


class TestLUT3D_Jakob2019(unittest.TestCase):
    """
    Define :class:`colour.recovery.jakob2019.LUT3D_Jakob2019`
    definition unit tests methods.
    """

    def setUp(self):
        """Initialise the common tests attributes."""

        self._shape = SPECTRAL_SHAPE_JAKOB2019
        self._cmfs, self._sd_D65 = handle_spectral_arguments(
            shape_default=self._shape
        )
        self._XYZ_D65 = sd_to_XYZ(self._sd_D65)
        self._xy_D65 = XYZ_to_xy(self._XYZ_D65)

        self._RGB_colourspace = RGB_COLOURSPACE_sRGB

        self._temporary_directory = tempfile.mkdtemp()

        self._LUT = LUT3D_Jakob2019()
        self._LUT.generate(self._RGB_colourspace, self._cmfs, self._sd_D65, 5)

        self._path = os.path.join(
            self._temporary_directory, "Test_Jakob2019.coeff"
        )
        self._LUT.write(self._path)

    def tearDown(self):
        """After tests actions."""

        shutil.rmtree(self._temporary_directory)

    def test_required_attributes(self):
        """Test the presence of required attributes."""

        required_attributes = (
            "size",
            "lightness_scale",
            "coefficients",
            "interpolator",
        )

        for attribute in required_attributes:
            self.assertIn(attribute, dir(LUT3D_Jakob2019))

    def test_required_methods(self):
        """Test the presence of required methods."""

        required_methods = (
            "__init__",
            "generate",
            "RGB_to_coefficients",
            "RGB_to_sd",
            "read",
            "write",
        )

        for method in required_methods:
            self.assertIn(method, dir(LUT3D_Jakob2019))

    def test_size(self):
        """Test :attr:`colour.recovery.jakob2019.LUT3D_Jakob2019.size` property."""

        self.assertEqual(self._LUT.size, 5)

    def test_lightness_scale(self):
        """
        Test :attr:`colour.recovery.jakob2019.LUT3D_Jakob2019.lightness_scale`
        property.
        """

        np.testing.assert_almost_equal(
            self._LUT.lightness_scale,
            np.array(
                [0.00000000, 0.06561279, 0.50000000, 0.93438721, 1.00000000]
            ),
            decimal=7,
        )

    def test_coefficients(self):
        """
        Test :attr:`colour.recovery.jakob2019.LUT3D_Jakob2019.coefficients`
        property.
        """

        self.assertTupleEqual(self._LUT.coefficients.shape, (3, 5, 5, 5, 3))

    def test_LUT3D_Jakob2019(self):
        """
        Test the entirety of the
        :class:`colour.recovery.jakob2019.LUT3D_Jakob2019`class.
        """

        LUT = LUT3D_Jakob2019()
        LUT.read(self._path)

        for RGB in [
            np.array([1, 0, 0]),
            np.array([0, 1, 0]),
            np.array([0, 0, 1]),
            zeros(3),
            full(3, 0.5),
            ones(3),
        ]:
            XYZ = RGB_to_XYZ(
                RGB,
                self._RGB_colourspace.whitepoint,
                self._xy_D65,
                self._RGB_colourspace.matrix_RGB_to_XYZ,
            )
            Lab = XYZ_to_Lab(XYZ, self._xy_D65)

            recovered_sd = LUT.RGB_to_sd(RGB)
            recovered_XYZ = (
                sd_to_XYZ(recovered_sd, self._cmfs, self._sd_D65) / 100
            )
            recovered_Lab = XYZ_to_Lab(recovered_XYZ, self._xy_D65)

            error = delta_E_CIE1976(Lab, recovered_Lab)

            if error > 2 * JND_CIE1976 / 100:  # pragma: no cover
                self.fail(
                    f"Delta E for RGB={RGB} in colourspace "
                    f"{self._RGB_colourspace.name} is {error}!"
                )

    def test_raise_exception_RGB_to_coefficients(self):
        """
        Test :func:`colour.recovery.jakob2019.LUT3D_Jakob2019.\
RGB_to_coefficients` method raised exception.
        """

        LUT = LUT3D_Jakob2019()

        self.assertRaises(RuntimeError, LUT.RGB_to_coefficients, np.array([]))

    def test_raise_exception_read(self):
        """
        Test :func:`colour.recovery.jakob2019.LUT3D_Jakob2019.read` method
        raised exception.
        """

        self.assertRaises(ValueError, self._LUT.read, __file__)


if __name__ == "__main__":
    unittest.main()



"""Defines the unit tests for the :mod:`colour.temperature.ohno2013` module."""

import numpy as np
import unittest
from itertools import permutations

from colour.colorimetry import MSDS_CMFS
from colour.temperature import CCT_to_uv_Ohno2013, uv_to_CCT_Ohno2013
from colour.temperature.ohno2013 import (
    planckian_table,
    planckian_table_minimal_distance_index,
)
from colour.utilities import ignore_numpy_errors

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestPlanckianTable",
    "TestPlanckianTableMinimalDistanceIndex",
    "Testuv_to_CCT_Ohno2013",
    "TestCCT_to_uv_Ohno2013",
]

PLANCKIAN_TABLE = np.array(
    [
        [1000.00000000, 0.44796288, 0.35462962, 0.25373557],
        [1001.11111111, 0.44770303, 0.35465214, 0.25348315],
        [1002.22222222, 0.44744348, 0.35467461, 0.25323104],
        [1003.33333333, 0.44718423, 0.35469704, 0.25297924],
        [1004.44444444, 0.44692529, 0.35471942, 0.25272774],
        [1005.55555556, 0.44666666, 0.35474175, 0.25247656],
        [1006.66666667, 0.44640833, 0.35476404, 0.25222569],
        [1007.77777778, 0.44615030, 0.35478628, 0.25197512],
        [1008.88888889, 0.44589258, 0.35480848, 0.25172487],
        [1010.00000000, 0.44563516, 0.35483063, 0.25147492],
    ]
)


class TestPlanckianTable(unittest.TestCase):
    """
    Define :func:`colour.temperature.ohno2013.planckian_table` definition
    unit tests methods.
    """

    def test_planckian_table(self):
        """Test :func:`colour.temperature.ohno2013.planckian_table` definition."""

        np.testing.assert_almost_equal(
            [
                (x.Ti, x.ui, x.vi, x.di)
                for x in planckian_table(
                    np.array([0.1978, 0.3122]),
                    MSDS_CMFS["CIE 1931 2 Degree Standard Observer"],
                    1000,
                    1010,
                    10,
                )
            ],
            PLANCKIAN_TABLE,
        )


class TestPlanckianTableMinimalDistanceIndex(unittest.TestCase):
    """
    Define :func:`colour.temperature.ohno2013.\
planckian_table_minimal_distance_index` definition unit tests methods.
    """

    def test_planckian_table_minimal_distance_index(self):
        """
        Test :func:`colour.temperature.ohno2013.\
planckian_table_minimal_distance_index` definition.
        """

        self.assertEqual(
            planckian_table_minimal_distance_index(
                planckian_table(
                    np.array([0.1978, 0.3122]),
                    MSDS_CMFS["CIE 1931 2 Degree Standard Observer"],
                    1000,
                    1010,
                    10,
                )
            ),
            9,
        )


class Testuv_to_CCT_Ohno2013(unittest.TestCase):
    """
    Define :func:`colour.temperature.ohno2013.uv_to_CCT_Ohno2013` definition
    unit tests methods.
    """

    def test_uv_to_CCT_Ohno2013(self):
        """
        Test :func:`colour.temperature.ohno2013.uv_to_CCT_Ohno2013`
        definition.
        """

        np.testing.assert_almost_equal(
            uv_to_CCT_Ohno2013(np.array([0.1978, 0.3122])),
            np.array([6507.47380460, 0.00322335]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            uv_to_CCT_Ohno2013(np.array([0.4328, 0.2883])),
            np.array([1041.68315360, -0.06737802]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            uv_to_CCT_Ohno2013(np.array([0.2927, 0.2722]), iterations=4),
            np.array([2452.15316417, -0.08437064]),
            decimal=7,
        )

    def test_n_dimensional_uv_to_CCT_Ohno2013(self):
        """
        Test :func:`colour.temperature.ohno2013.uv_to_CCT_Ohno2013` definition
        n-dimensional arrays support.
        """

        uv = np.array([0.1978, 0.3122])
        CCT_D_uv = uv_to_CCT_Ohno2013(uv)

        uv = np.tile(uv, (6, 1))
        CCT_D_uv = np.tile(CCT_D_uv, (6, 1))
        np.testing.assert_almost_equal(
            uv_to_CCT_Ohno2013(uv), CCT_D_uv, decimal=7
        )

        uv = np.reshape(uv, (2, 3, 2))
        CCT_D_uv = np.reshape(CCT_D_uv, (2, 3, 2))
        np.testing.assert_almost_equal(
            uv_to_CCT_Ohno2013(uv), CCT_D_uv, decimal=7
        )

    @ignore_numpy_errors
    def test_nan_uv_to_CCT_Ohno2013(self):
        """
        Test :func:`colour.temperature.ohno2013.uv_to_CCT_Ohno2013` definition
        nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=2))
        for case in cases:
            uv = np.array(case)
            uv_to_CCT_Ohno2013(uv)


class TestCCT_to_uv_Ohno2013(unittest.TestCase):
    """
    Define :func:`colour.temperature.ohno2013.CCT_to_uv_Ohno2013` definition
    unit tests methods.
    """

    def test_CCT_to_uv_Ohno2013(self):
        """
        Test :func:`colour.temperature.ohno2013.CCT_to_uv_Ohno2013`
        definition.
        """

        np.testing.assert_almost_equal(
            CCT_to_uv_Ohno2013(np.array([6507.47380460, 0.00322335])),
            np.array([0.19779997, 0.31219997]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            CCT_to_uv_Ohno2013(np.array([1041.68315360, -0.06737802])),
            np.array([0.43279885, 0.28830013]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            CCT_to_uv_Ohno2013(np.array([2452.15316417, -0.08437064])),
            np.array([0.29247364, 0.27215157]),
            decimal=7,
        )

    def test_n_dimensional_CCT_to_uv_Ohno2013(self):
        """
        Test :func:`colour.temperature.ohno2013.CCT_to_uv_Ohno2013` definition
        n-dimensional arrays support.
        """

        CCT_D_uv = np.array([6507.47380460, 0.00322335])
        uv = CCT_to_uv_Ohno2013(CCT_D_uv)

        CCT_D_uv = np.tile(CCT_D_uv, (6, 1))
        uv = np.tile(uv, (6, 1))
        np.testing.assert_almost_equal(
            CCT_to_uv_Ohno2013(CCT_D_uv), uv, decimal=7
        )

        CCT_D_uv = np.reshape(CCT_D_uv, (2, 3, 2))
        uv = np.reshape(uv, (2, 3, 2))
        np.testing.assert_almost_equal(
            CCT_to_uv_Ohno2013(CCT_D_uv), uv, decimal=7
        )

    @ignore_numpy_errors
    def test_nan_CCT_to_uv_Ohno2013(self):
        """
        Test :func:`colour.temperature.ohno2013.CCT_to_uv_Ohno2013` definition
        nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=2))
        for case in cases:
            CCT_D_uv = np.array(case)
            CCT_to_uv_Ohno2013(CCT_D_uv)


if __name__ == "__main__":
    unittest.main()
